from epyk.core.data.DataClass import DataClass


class SearchPanes(DataClass):

  class DtOps(DataClass):
    @property
    def paging(self):
      return self._attrs['paging']

    @paging.setter
    def paging(self, val):
      self._attrs['paging'] = val

    @property
    def pagingType(self):
      return self._attrs['pagingType']

    @pagingType.setter
    def pagingType(self, val):
      self._attrs['pagingType'] = val

    @property
    def searching(self):
      return self._attrs['searching']

    @searching.setter
    def searching(self, val):
      self._attrs['searching'] = val

  class Panes(DataClass):

    class Options(DataClass):
      @property
      def label(self):
        return self._attrs['label']

      @label.setter
      def label(self, val):
        self._attrs['label'] = val


    class DtOps(DataClass):
      @property
      def paging(self):
        return self._attrs['paging']

      @paging.setter
      def paging(self, val):
        self._attrs['paging'] = val

      @property
      def pagingType(self):
        return self._attrs['pagingType']

      @pagingType.setter
      def pagingType(self, val):
        self._attrs['pagingType'] = val

      @property
      def searching(self):
        return self._attrs['searching']

      @searching.setter
      def searching(self, val):
        self._attrs['searching'] = val


    @property
    def className(self):
      return self._attrs['className']

    @className.setter
    def className(self, val):
      self._attrs['className'] = val

    @property
    def header(self):
      return self._attrs['header']

    @header.setter
    def header(self, val):
      self._attrs['header'] = val

    @property
    def panes_options(self):
      return self.sub_data_enum('options', self.Options)

    @property
    def dtOpts(self):
      return self.sub_data('dtOpts', self.DtOps)


  @property
  def cascadePanes(self):
    """
    Description:
    -----------
    As standard, SearchPanes will display all of the options for that column throughout its operation.
    When searchPanes.cascadePanes is set to true, the panes will remove options which are no longer present in the currently displayed DataTable.
    The cascade action will occur when a selection or a deselection is made in one of the panes.

    This is useful when dealing with large data sets with many different options as they are narrowed down accordingly as selections proceed.

    Related Pages:

			https://datatables.net/reference/option/searchPanes.cascadePanes
    """
    return self._attrs['cascadePanes']

  @cascadePanes.setter
  def cascadePanes(self, val):
    self._attrs['cascadePanes'] = val

  @property
  def clear(self):
    """
    Description:
    -----------
    As standard, SearchPanes will display the buttons which allow selections to be cleared in individual panes and across all panes.
    When searchPanes.clear is set to false, the clear buttons will not be present.

    The searchPanes.clear functionality is useful when wanting to quickly deselect all of the options in a pane.
    It is useful to be able to disable these buttons, for example on smaller tables with less options it may be preferrable to employ a custom user interface.

    Related Pages:

			https://datatables.net/reference/option/searchPanes.clear
    """
    return self._attrs['clear']

  @clear.setter
  def clear(self, val):
    self._attrs['clear'] = val

  @property
  def columns(self):
    """
    Description:
    -----------
    As standard, SearchPanes will consider all of the columns when creating searchPanes.
    When searchPanes.columns is defined with an array of numbers, then only the columns with those indexes will be considered when creating panes.
    Note they still may not be displayed based upon the calculations of searchPanes.threshold or columns.searchPanes.threshold.

    The searchPanes.columns functionality is useful when wanting to restrict which columns will be considered when displaying panes.
    This may result in faster load times as there is not as much validation to be done.

    Related Pages:

			https://datatables.net/reference/option/searchPanes.columns
    """
    return self._attrs['columns']

  @columns.setter
  def columns(self, val):
    self._attrs['columns'] = val

  @property
  def controls(self):
    """
    Description:
    -----------
    As standard, SearchPanes will be displayed with the control buttons included in the interface with searching enabled.
    However if the value of searchPanes.controls is set to false then the control buttons will no longer be displayed and searching will be disabled in all of the panes.

    The control buttons can be hidden and searching disabled for individual panes by the columns.searchPanes.controls option.

    Related Pages:

			https://datatables.net/reference/option/searchPanes.controls
    """
    return self._attrs['controls']

  @controls.setter
  def controls(self, val):
    self._attrs['controls'] = val

  @property
  def dataLength(self):
    """
    Description:
    -----------
    This is useful as it prevents the searchPanes cells becoming mishapen and overpopulated when the table contains long strings.

    Related Pages:

			https://datatables.net/reference/option/searchPanes.dataLength
    """
    return self._attrs['dataLength']

  @dataLength.setter
  def dataLength(self, val):
    self._attrs['dataLength'] = val

  @property
  def emptyMessage(self):
    """
    Description:
    -----------
    Empty cells will be represented in the panes by the option searchPanes.emptyMessage. This is useful as it's a more user-friendly way than just having a blank cell.

    Related Pages:

			https://datatables.net/reference/option/searchPanes.emptyMessage
    """
    return self._attrs['emptyMessage']

  @emptyMessage.setter
  def emptyMessage(self, val):
    self._attrs['emptyMessage'] = val

  @property
  def hideCount(self):
    """
    Description:
    -----------
    As standard, SearchPanes will be displayed with the count column included in the table.
    However if the value of searchPanes.hideCount is set to true then the count column won't be displayed in any of the panes.

    Related Pages:

			https://datatables.net/reference/option/searchPanes.hideCount
    """
    return self._attrs['hideCount']

  @hideCount.setter
  def hideCount(self, val):
    self._attrs['hideCount'] = val

  @property
  def layout(self):
    """
    Description:
    -----------
    By setting the searchPanes.layout parameter to columns-3, the panes will be displayed in 3 columns.

    There are 6 sizes supported from 1 pane in a row to 6 panes in a row. Altering the number at the end of columns- will alter the number of panes on a row.

    Related Pages:

			https://datatables.net/reference/option/searchPanes.layout
    """
    return self._attrs['layout']

  @layout.setter
  def layout(self, val):
    self._attrs['layout'] = val

  @property
  def orderable(self):
    """
    Description:
    -----------
    As standard, SearchPanes will be displayed with the ordering buttons included in the interface, allowing the user to change the order of the selections in each pane.
    However if the value of searchPanes.orderable is set to false then the ordering buttons will no longer be displayed in any of the panes.
    Related Pages:

			https://datatables.net/reference/option/searchPanes.orderable
    """
    return self._attrs['orderable']

  @orderable.setter
  def layout(self, val):
    self._attrs['orderable'] = val

  @property
  def orderable(self):
    """
    Description:
    -----------
    As standard, SearchPanes will be displayed with the ordering buttons included in the interface, allowing the user to change the order of the selections in each pane.
    However if the value of searchPanes.orderable is set to false then the ordering buttons will no longer be displayed in any of the panes.
    Related Pages:

			https://datatables.net/reference/option/searchPanes.orderable
    """
    return self._attrs['orderable']

  @orderable.setter
  def orderable(self, val):
    self._attrs['orderable'] = val

  @property
  def threshold(self):
    """
    Description:
    -----------
    Set the minimum number of unique values needed in the columns to display that pane.
    As standard, columns must have a uniqueness ratio of 0.6. This is the ratio of different values throughout the column to the number of total rows.
    If all rows are unique, this will be 1; as the value drops towards 0, this means values are more often repeated.
    --------------
    https://datatables.net/reference/option/searchPanes.threshold
    """
    return self._attrs['threshold']

  @threshold.setter
  def threshold(self, val):
    self._attrs['threshold'] = val

  @property
  def viewTotal(self):
    """
    Description:
    -----------
    Update the count column when searching to show visible count.
    By setting the searchPanes.viewTotal parameter to true, the message displayed in the count column will change when a search is applied to the DataTable.

    As standard, SearchPanes will set the count column to display the total number of different values visible in the column when no filtering is applied.
    When filtering is active the count column will display the number of entries that satisfy this condition currently on display followed in brackets by the total in the table which satisfy the condition.
    --------------
    https://datatables.net/reference/option/searchPanes.viewTotal
    """
    return self._attrs['viewTotal']

  @viewTotal.setter
  def viewTotal(self, val):
    self._attrs['viewTotal'] = val

  @property
  def dtOpts(self):
    """
    Description:
    -----------
    This is useful as it prevents the searchPanes cells becoming mishapen and overpopulated when the table contains long strings.

    Related Pages:

			https://datatables.net/reference/option/searchPanes.dtOpts
    """
    return self.sub_data('dtOps', self.DtOps)

  @property
  def panes(self):
    """
    Description:
    -----------
    Define custom panes to filter across all columns.
    --------------
    https://datatables.net/reference/option/searchPanes.panes
    """
    return self.sub_data_enum('panes', self.Panes)



"""

TODO: Split the component_properies class attribute into two decorators
@js_option =>  self._report._jsStyles
@html_option =>  self._attrs
"""

import sys

from epyk.core.data.DataClass import DataClass


class Options(DataClass):
  component_properies = ()

  def __init__(self, report, attrs=None, options=None):
    super(Options, self).__init__(report, attrs, options)
    self.js_type = {}
    if attrs is not None:
      for k, v in attrs.items():
        if hasattr(self, k):
          setattr(self, k, v)
        else:
          self._report._jsStyles[k] = v

  def _config_get(self, dflt=None, name=None):
    """
    Description:
    ------------
    Get the option attribute to be added on the Javascript side during the component build

    Attributes:
    ----------
    :param name: String. The attribute name
    """
    return self._report._jsStyles.get(name or sys._getframe().f_back.f_code.co_name, dflt)

  def _config(self, value, name=None):
    """
    Description:
    ------------
    Set the option attribute to be added on the Javascript side during the component build

    Attributes:
    ----------
    :param value: Object. The value for the name
    :param name: String. The attribute name
    """
    self._report._jsStyles[name or sys._getframe().f_back.f_code.co_name] = value

  def _config_group_get(self, group, dflt=None, name=None):
    """
    Description:
    ------------
    Get second level configuration options

    Attributes:
    ----------
    :param group: String. The group attribute name
    :param name: String. The attribute name
    """
    return self._report._jsStyles.get(group, {}).get(name or sys._getframe().f_back.f_code.co_name, dflt)

  def _config_group(self, group, value, name=None):
    """
    Description:
    ------------
    Set second level configuration options

    Attributes:
    ----------
    :param group: String. The group name
    :param value: Object. The value for the name
    :param name: String. The attribute name
    """
    if not group in self._report._jsStyles:
      self._report._jsStyles[group] = {}
    self._report._jsStyles[group][name or sys._getframe().f_back.f_code.co_name] = value

  def isJsContent(self, property_name):
    """
    Description:
    ------------
    Check if the content of a property is defined to always be a JavaScript fragment.
    Thus the frameowrk will not convert it to a Json content

    Attributes:
    ----------
    :param property_name:
    """
    return self.js_type.get(property_name, False)

  @property
  def managed(self):
    """
    Description:
    ------------
    """
    return self.get(True)

  @managed.setter
  def managed(self, bool):
    self.set(bool)

  def details(self):
    """
    Description:
    ------------
    Retrieve the defined properties details
    """
    prop_details = []
    for prop_name in self.component_properies:
      prop = getattr(self, prop_name)
      prop_details.append({"name": prop_name, 'value': prop, 'doc':  getattr(self.__class__, prop_name).__doc__})
    return prop_details



"""

Related Pages:

		https://naver.github.io/billboard.js/release/latest/doc/Chart.html#data%25E2%2580%25A4colors
"""

from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects
from epyk.core.js.packages import JsPackage


class C3Legend(object):
  def __init__(self, varName):
    self.varName = varName

  def show(self, targetIds=None):
    """
    Description:
    -----------
    Show legend for each target.

    https://c3js.org/reference.html#api-legend-show

    :param targetIds:
    """
    if targetIds is None:
      return "%s.show()" % self.varName

    targetIds = JsUtils.jsConvertData(targetIds, None)
    return "%s.show(%s)" % (self.varName, targetIds)

  def hide(self, targetIds=None):
    """
    Description:
    -----------
    Show legend for each target.

    https://c3js.org/reference.html#api-legend-show

    :param targetIds:
    """
    if targetIds is None:
      return "%s.hide()" % self.varName

    targetIds = JsUtils.jsConvertData(targetIds, None)
    return "%s.hide(%s)" % (self.varName, targetIds)


class C3Data(object):
  def __init__(self, varName):
    self.varName = varName

  def axes(self, targetIds=None):
    """
    Description:
    -----------
    Get and set axes of the data loaded in the chart.

    https://c3js.org/reference.html#api-data-axes

    Attributes:
    ----------
    :param targetIds:
    """
    if targetIds is None:
      return "%s.axes()" % self.varName

    targetIds = JsUtils.jsConvertData(targetIds, None)
    return "%s.axes(%s)" % (self.varName, targetIds)

  def values(self, targetIds=None):
    """
    Description:
    -----------
    Get values of the data loaded in the chart.

    https://c3js.org/reference.html#api-data-values

    Attributes:
    ----------
    :param targetIds:
    """
    if targetIds is None:
      return "%s.values()" % self.varName

    targetIds = JsUtils.jsConvertData(targetIds, None)
    return "%s.values(%s)" % (self.varName, targetIds)

  def colors(self, targetIds=None):
    """
    Description:
    -----------
    Get and set colors of the data loaded in the chart.

    https://c3js.org/reference.html#api-data-colors

    Attributes:
    ----------
    :param targetIds:
    """
    if targetIds is None:
      return "%s.colors()" % self.varName

    targetIds = JsUtils.jsConvertData(targetIds, None)
    return "%s.colors(%s)" % (self.varName, targetIds)

  def names(self, targetIds=None):
    """
    Description:
    -----------
    Get and set names of the data loaded in the chart.

    https://c3js.org/reference.html#api-data-names

    Attributes:
    ----------
    :param targetIds:
    """
    if targetIds is None:
      return "%s.names()" % self.varName

    targetIds = JsUtils.jsConvertData(targetIds, None)
    return "%s.names(%s)" % (self.varName, targetIds)

  def show(self, targetIds=None):
    """
    Description:
    -----------
    Show legend for each target.

    https://c3js.org/reference.html#api-legend-show

    Attributes:
    ----------
    :param targetIds:
    """
    if targetIds is None:
      return "%s.show()" % self.varName

    targetIds = JsUtils.jsConvertData(targetIds, None)
    return "%s.show(%s)" % (self.varName, targetIds)

  def hide(self, targetIds=None):
    """
    Description:
    -----------
    Show legend for each target.

    https://c3js.org/reference.html#api-legend-show

    Attributes:
    ----------
    :param targetIds:
    """
    if targetIds is None:
      return "%s.hide()" % self.varName

    targetIds = JsUtils.jsConvertData(targetIds, None)
    return "%s.hide(%s)" % (self.varName, targetIds)


class C3(JsPackage):
  lib_alias = {'js': "c3", 'css': 'c3'}

  def __init__(self, src, varName, setVar=False, report=None):
    super(C3, self).__init__(src=src, varName=varName, selector=varName, setVar=setVar)
    self._report = report

  @property
  def legend(self):
    return C3Legend("%s.legend" % self.src.htmlCode)

  @property
  def value(self):
    return JsObjects.JsObjects.get("{'%s': arguments[0], timestamp: Date.now(), offset: new Date().getTimezoneOffset()}" % self.src.htmlCode)

  @property
  def content(self):
    return JsObjects.JsNumber.JsNumber("arguments[0].value")

  @property
  def label(self):
    return JsObjects.JsString.JsString("arguments[0].name", isPyData=False)

  def load(self):
    """

    :return:
    """

  def show(self, targetIds=None, options=None):
    """
    Description:
    -----------
    This API shows specified targets.

    You can specify multiple targets by giving an array that includes id as String. If no argument is given, all of targets will be shown.

    https://c3js.org/reference.html#api-show

    Attributes:
    ----------
    :param targetIds:
    :param options:
    """
    if targetIds is None and options is None:
      return JsObjects.JsVoid("%s.show()" % (self._selector))

    targetIds = JsUtils.jsConvertData(targetIds, None)
    options = JsUtils.jsConvertData(options, None)
    return JsObjects.JsVoid("%s.show(%s, %s)" % (self._selector, targetIds, options))

  def toggle(self, targetIds=None, options=None):
    """
    Description:
    -----------
    This API shows specified targets.

    You can specify multiple targets by giving an array that includes id as String. If no argument is given, all of targets will be shown.

    https://c3js.org/reference.html#api-show

    Attributes:
    ----------
    :param targetIds:
    :param options:
    """
    if targetIds is None and options is None:
      return JsObjects.JsVoid("%s.toggle()" % (self._selector))

    targetIds = JsUtils.jsConvertData(targetIds, None)
    options = JsUtils.jsConvertData(options, None)
    return JsObjects.JsVoid("%s.toggle(%s, %s)" % (self._selector, targetIds, options))

  @property
  def data(self):
    """
    Description:
    -----------
    Get data loaded in the chart.

    https://c3js.org/reference.html#api-data
    """
    return

  def hide(self, targetIds=None, options=None):
    """
    Description:
    -----------
    This API hides specified targets.

    You can specify multiple targets by giving an array that includes id as String. If no argument is given, all of targets will be hidden.

    https://c3js.org/reference.html#api-hide

    Attributes:
    ----------
    :param targetIds:
    :param options:
    """
    if targetIds is None and options is None:
      return JsObjects.JsVoid("%s.show()" % (self._selector))

    targetIds = JsUtils.jsConvertData(targetIds, None)
    options = JsUtils.jsConvertData(options, None)
    return JsObjects.JsVoid("%s.show(%s, %s)" % (self._selector, targetIds, options))

  def flow(self, args):
    """
    Description:
    -----------
    Flow data to the chart.

    By this API, you can append new data points to the chart.

    https://c3js.org/reference.html#api-flow

    Attributes:
    ----------
    :param args:
    """
    args = JsUtils.jsConvertData(args, None)
    return JsObjects.JsVoid("%s.flow(%s)" % (self._selector, args))

  def transform(self, type, targetIds=None):
    """
    Description:
    -----------
    Change the type of the chart.

    https://c3js.org/reference.html#api-transform

    :param args:
    """
    if targetIds is None:
      type = JsUtils.jsConvertData(type, None)
      return JsObjects.JsVoid("%s.transform(%s)" % (self._selector, type))

    targetIds = JsUtils.jsConvertData(targetIds, None)
    return JsObjects.JsVoid("%s.transform(%s, %s)" % (self._selector, type, targetIds))

  def load(self, args):
    """
    Description:
    -----------
    Load data to the chart.

    You can specify multiple targets by giving an array that includes id as String. If no argument is given, all of targets will be toggles.

    https://c3js.org/reference.html#api-load

    Attributes:
    ----------
    :param args:
    """
    if isinstance(args, list):
      if not isinstance(args[0], list):
        args = {"columns": [args]}
      else:
        args = {"columns": args}
    args = JsUtils.jsConvertData(args, None)
    return JsObjects.JsVoid("%s.load(%s)" % (self._selector, args))

  def unload(self, args):
    """
    Description:
    -----------
    Unload data to the chart.

    You can specify multiple targets by giving an array that includes id as String. If no argument is given, all of targets will be toggles.

    https://c3js.org/reference.html#api-unload

    Attributes:
    ----------
    :param args:
    """
    if isinstance(args, list):
      args = {"ids": args}
    args = JsUtils.jsConvertData(args, None)
    return JsObjects.JsVoid("%s.unload(%s)" % (self._selector, args))

  def resise(self, size):
    """
    Description:
    -----------
    Resize the chart.

    https://c3js.org/reference.html#api-resize
    """
    size = JsUtils.jsConvertData(size, None)
    return JsObjects.JsVoid("%s.resise(%s)" % (self._selector, size))

  def flush(self):
    """
    Description:
    -----------
    Force to redraw.

    https://c3js.org/reference.html#api-flush
    """
    return JsObjects.JsVoid("%s.flush()" % self._selector)

  def destroy(self):
    """
    Description:
    -----------
    Reset the chart object and remove element and events completely.

    https://c3js.org/reference.html#api-flush
    """
    return JsObjects.JsVoid("%s.destroy()" % self._selector)

  def xgrids(self, grids):
    """
    Description:
    -----------
    Update x grid lines.
    """
    grids = JsUtils.jsConvertData(grids, None)
    return JsObjects.JsVoid("%s.xgrids(%s)" % (self._selector, grids))

  def select(self, ids, indices, resetOthers):
    """
    Description:
    -----------
    Change data point state to selected.

    By this API, you can select data points. To use this API, data.selection.enabled needs to be set true.

    https://c3js.org/reference.html#api-select
    """
    ids = JsUtils.jsConvertData(ids, None)
    indices = JsUtils.jsConvertData(indices, None)
    resetOthers = JsUtils.jsConvertData(resetOthers, None)
    return JsObjects.JsVoid("%s.select(%s, %s, %s)" % (self._selector, ids, indices, resetOthers))

  def unselected(self, ids, indices):
    """
    Description:
    -----------
    Change data point state to unselected.

    By this API, you can unselect data points. To use this API, data.selection.enabled needs to be set true.

    https://c3js.org/reference.html#api-select
    """
    ids = JsUtils.jsConvertData(ids, None)
    indices = JsUtils.jsConvertData(indices, None)
    return JsObjects.JsVoid("%s.select(%s, %s)" % (self._selector, ids, indices))

  def selected(self, ids):
    """
    Description:
    -----------
    Get selected data points.

    By this API, you can get selected data points information. To use this API, data.selection.enabled needs to be set true.

    https://c3js.org/reference.html#api-selected
    """
    ids = JsUtils.jsConvertData(ids, None)
    return JsObjects.JsVoid("%s.selected(%s)" % (self._selector, ids))

  def zoom(self, domain):
    """
    Description:
    -----------
    Zoom by giving x domain.

    https://c3js.org/reference.html#api-zoom
    """
    domain = JsUtils.jsConvertData(domain, None)
    return JsObjects.JsVoid("%s.zoom(%s)" % (self._selector, domain))

  def unzoom(self):
    """
    Description:
    -----------
    Unzoom to the original domain.

    https://c3js.org/reference.html#api-unzoom
    """
    return JsObjects.JsVoid("%s.unzoom()" % self._selector)




from epyk.core.css.styles.classes import CssStyle

from epyk.core.css import Defaults as Defaults_css


class CssTabulator(CssStyle.Style):

  def customize(self):
    self.css({'border': '1px solid %s !IMPORTANT' % self.rptObj.theme.greys[3],
              'font-size': '%s%s' % (Defaults_css.Font.header_size, Defaults_css.Font.unit),
              'background-color': self.rptObj.theme.greys[2], 'font-family': Defaults_css.Font.family})


class CssTabulatorFooter(CssStyle.Style):
  classname = "tabulator-footer"

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'border': 'none', 'background': self.rptObj.theme.greys[0]}, important=True)


class CssTabulatorHeader(CssStyle.Style):
  classname = "tabulator-header"

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[0], 'border': 'none', 'background': 'none'}, important=True)


class CssTabulatorHeaders(CssStyle.Style):
  classname = 'tabulator-headers'

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[0], 'border': 'none', 'background': 'none'}, important=True)


class CssTabulatorSelected(CssStyle.Style):
  classname = 'tabulator-selected'

  def customize(self):
    self.css({'border': '1px solid %s' % self.rptObj.theme.success[1], 'color': 'black',
              'background': self.rptObj.theme.success[0]}, important=True)


class CssTabulatorCol(CssStyle.Style):
  classname = 'tabulator-col'

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[0], 'border': '1px solid %s' % self.rptObj.theme.greys[-1],
              'background': self.rptObj.theme.greys[-1], 'padding': '4px 0', "text-align": 'center'}, important=True)


class CssTabulatorColContent(CssStyle.Style):
  classname = 'tabulator-col-content'

  def customize(self):
    self.css({'color': "#DEDEDE", 'border': '1px solid #212121', 'background': "#212121"})


class CssTabulatorFooterPagination(CssStyle.Style):
  classname = 'tabulator-page:not(.disabled)'

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'border': 'none', 'background': self.rptObj.theme.greys[0]}, important=True)
    self.hover.css({'color': self.rptObj.theme.colors[0], 'background': self.rptObj.theme.colors[-1]}, important=True)


class CssTabulatorGroups(CssStyle.Style):
  _attrs = {'text-align': 'center !IMPORTANT'}
  classname = 'tabulator-group'


class CssTabulatorEvenRow(CssStyle.Style):
  classname = 'tabulator-row-even'

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'border': "1px solid %s" % self.rptObj.theme.colors[1],
              'background-color': self.rptObj.theme.colors[1]}, important=True)


class CssTabulatorOddRow(CssStyle.Style):
  classname = 'tabulator-row-odd'

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'border': "1px solid %s" % self.rptObj.theme.greys[0],
              'background-color': self.rptObj.theme.greys[0]}, important=True)


class CssTabulatorRow(CssStyle.Style):
  classname = 'tabulator-row'

  def customize(self):
    self.hover.css({'color': self.rptObj.theme.greys[-1], 'border': "1px solid %s" % self.rptObj.theme.success[1],
                    'background-color': self.rptObj.theme.colors[3]}, important=True)


class CssTabulatorCell(CssStyle.Style):
  classname = 'tabulator-cell'

  def customize(self):
    self.css({'border-right': "1px solid %s" % self.rptObj.theme.greys[3]}, important=True)


class CssTabulatorTreeControl(CssStyle.Style):
  classname = 'tabulator-data-tree-control'

  def customize(self):
    self.css({'border-color': self.rptObj.theme.greys[-1]}, important=True)


class CssTabulatorTreeControlExpand(CssStyle.Style):
  classname = 'tabulator-data-tree-control-expand'

  def customize(self):
    self.css({'background': self.rptObj.theme.greys[-1]}, important=True)




from epyk.core.data.DataClass import DataClass


class ColReorder(DataClass):

  def activate(self):
    """
    Description:
    -----------
    ColReorder provides the option for end users to reorder columns in a DataTable by click and drag, or for yourself, the developer using DataTable, through the API.

    Related Pages:

			https://datatables.net/reference/option/colReorder
    """
    self.realtime = True
    return self

  @property
  def enable(self):
    """
    Description:
    -----------
    It can be useful to disable ColReorder's user input controls at certain times, depending on the state of your application. This option provides that ability when the table is initially created, while the colReorder.enable() and colReorder.disable() methods provide the option to enabling the user interaction after the table has been created.

    Related Pages:

			https://datatables.net/reference/option/colReorder.enable
    """
    return self._attrs["enable"]

  @enable.setter
  def enable(self, val):
    self._attrs["enable"] = val

  @property
  def fixedColumnsLeft(self):
    """
    Description:
    -----------
    When allowing reordering of columns in a table, you may often wish to disallow reordering of certain columns (for example locking an index, select or action column to the start of a table). This option provides that ability, locking columns counting from the left (colReorder.fixedColumnsRight provides the option to count from the right).

    Related Pages:

			https://datatables.net/reference/option/colReorder.fixedColumnsLeft
    """
    return self._attrs["fixedColumnsLeft"]

  @fixedColumnsLeft.setter
  def fixedColumnsLeft(self, val):
    self._attrs["fixedColumnsLeft"] = val

  @property
  def fixedColumnsRight(self):
    """
    Description:
    -----------
    When allowing reordering of columns in a table, you may often wish to disallow reordering of certain columns (for example locking an index, select or action column to the start of a table). This option provides that ability, locking columns counting from the right (colReorder.fixedColumnsLeft provides the option to count from the left).

    Related Pages:

			https://datatables.net/reference/option/colReorder.fixedColumnsRight
    """
    return self._attrs["fixedColumnsRight"]

  @fixedColumnsRight.setter
  def fixedColumnsRight(self, val):
    self._attrs["fixedColumnsRight"] = val

  @property
  def order(self):
    """
    Description:
    -----------
    This option provides the option to define a default order for the columns in a table. Typically you will wish to have the columns in the order defined in the HTML, or from state saving (stateSave), but if required, this option can be used to define an initial default order.

    Related Pages:

			https://datatables.net/reference/option/colReorder.order
    """
    return self._attrs["order"]

  @order.setter
  def order(self, val):
    self._attrs["order"] = val

  @property
  def realtime(self):
    """
    """
    return self._attrs["realtime"]

  @realtime.setter
  def realtime(self, val):
    self._attrs["realtime"] = val




from epyk.core.js import JsUtils

from epyk.core.js.packages import JsCrossFilter
from epyk.core.js.packages import JsDatatable
from epyk.core.js.packages import JsDc
from epyk.core.js.packages import JsRequire
from epyk.core.js.packages import JsTabulator


data = [
    {"name": "banana", "category": "fruit", "country": "Martinique", "outOfDateQuantity": 3, "quantity": 12},
    {"name": "apple", "category": "fruit", "country": "Spain", "outOfDateQuantity": 7, "quantity": 9},
    {"name": "tomato", "category": "vegetable", "country": "Spain", "outOfDateQuantity": 2, "quantity": 25}
  ]


# ------------------------------------------------------------------------------------------------------------------
#
f = JsUtils.JsFile("CrossFilter", path=r"../outs")
f.writeJs([
    JsCrossFilter.CrossFilter(data, 'cross1').dimension("quantity").filter([9, 13]),
    JsCrossFilter.CrossFilter(data, 'cross1').dimension("quantity").top(10),
    # JsCrossFilter.CrossFilter(data, 'cross1').dimension("category").group().reduceCount()
])
f.close()


# ------------------------------------------------------------------------------------------------------------------
#
f = JsUtils.JsFile("Datatable", path=r"../outs")
f.writeJs([
    JsDatatable.DatatableAPI().draw('page').clear().cell().data(),
    JsDatatable.DatatableAPI().draw('page').state(),
    JsDatatable.DatatableAPI().columns([1, 2]).search("%test").draw().jquery_nodes(),
    JsDatatable.DatatableAPI().columns([1, 2]).search("%test").draw()
])
f.close()


# ------------------------------------------------------------------------------------------------------------------
#
dcChart = JsDc.DC()
f = JsUtils.JsFile("DC", path=r"../outs")
f.writeJs([
    dcChart.width(23).height(23).toStr()
])
f.close()


# ------------------------------------------------------------------------------------------------------------------
#
#repObj = JsRequire.JsRequire(False)
#print(repObj.getImports(['socket.io', 'd3', 'bootstrap', 'tabulator', 'c3', 'datatables']))


# ------------------------------------------------------------------------------------------------------------------
#
tabulator = JsTabulator.Tabulator()
f = JsUtils.JsFile("Tabulator", path=r"../outs")
f.writeJs([
    tabulator.getRow(2).select()
])
f.close()



"""
CSS Style module for the common Div components
"""

from epyk.core.css.styles.classes import CssStyle

from epyk.core.css import Defaults as Defaults_css


class CssCommHeader(CssStyle.Style):
  _attrs = {'font-weight': 'bold', 'display': 'block', 'text-align': 'left'}
  _selectors = {'child': 'span'}

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'font-family': Defaults_css.Font.family,
              'font-size': '%s%s' % (Defaults_css.Font.size, Defaults_css.Font.unit)})


class CssCommInput(CssStyle.Style):
  _attrs = {'border': 'none', 'width': '95%', 'margin': 0, 'padding': '5px'}
  _focus = {'outline': 0}

  _selectors = {'child': 'input'}

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'background-color': self.rptObj.theme.greys[0],
              'border-bottom': '1px solid %s' % self.rptObj.theme.success[1], 'font-family': Defaults_css.Font.family,
              'font-size': '%s%s' % (Defaults_css.Font.size, Defaults_css.Font.unit)})


class CssContentEditable(CssStyle.Style):
  classname = False
  _selector = "[contenteditable]"

  def customize(self):
    self.focus.css({'outline': "1px solid %s" % self.rptObj.theme.success[1]})


class CssSpeechBubble(CssStyle.Style):
  classname = "speech-bubble"
  _attrs = {"border-radius": '4px', 'margin': '0 auto 10px', 'padding': '5px', 'position': 'relative'}


class CssSpeechBubbleContent(CssStyle.Style):
  classname = "speech-bubble-ds"
  _attrs = {"border": '1px solid #a7a7a7', 'border-radius': '4px', 'box-shadow': '4px 4px 0 rgba(0, 0, 0, 0.2)',
            'display': 'inline-block', 'margin': '0 auto 40px', 'padding': '0 5px', 'position': 'relative'}


class CssSpeechBubbleArrow(CssStyle.Style):
  classname = "speech-bubble-ds__arrow"
  _attrs = {"border-left": '21px solid transparent', "border-top": '20px solid rgba(0, 0, 0, 0.2)',
            'bottom': '-25px', 'position': 'absolute', 'right': '10px'}



"""
Description:
------------
Wrapper to the Number Javascript primitives

Most of the documentation in this module is coming from the website https://www.w3schools.com/jsref/jsref_number.asp

This module will ensure the conversion from Python to Javascript and from Javascript to Python.
Those intermediary object will be used as a bridge between the two languages.

They should be used in Js specific functions (expecting jsFncs or jsData) in order to be correctly converted.

Each function will have it is specific documentation and a simple example.

More examples can be found in the tests folder if needed.
If necessary the tests folder can be updated in order to catch some specific regressions
"""


from epyk.core.js.primitives import JsObject
from epyk.core.js import JsMaths


class JsNumber(JsObject.JsObject):
  _jsClass = "Number"

  @property
  def NEGATIVE_INFINITY(self):
    """
    Description:
    ------------
    The NEGATIVE_INFINITY property represents negative infinity.
    Negative infinity can be explained as something that is lower than any other number.

    Usage::

      jsObj.objects.number.get("MyNumber").NEGATIVE_INFINITY

    Related Pages:

			https://www.w3schools.com/jsref/jsref_negative_infinity.asp
    """
    return JsNumber("Number.NEGATIVE_INFINITY")

  @classmethod
  def POSITIVE_INFINITY(self):
    """
    Description:
    ------------
    The POSITIVE_INFINITY property represents positive infinity.
    Positive infinity can be explained as something that is higher than any other number.

    Usage::

      jsObj.objects.number.get("MyNumber").POSITIVE_INFINITY

    Related Pages:

			https://www.w3schools.com/jsref/jsref_positive_infinity.asp
    """
    return JsNumber("Number.POSITIVE_INFINITY")

  @classmethod
  def MAX_VALUE(self):
    """
    Description:
    ------------
    The MAX_VALUE property returns the largest number possible in JavaScript.

    This static property has a value of 1.7976931348623157e+308.

    Usage::

      jsObj.objects.number.get("MyNumber").MAX_VALUE

    Related Pages:

			https://www.w3schools.com/jsref/jsref_max_value.asp
    """
    return JsNumber("Number.MAX_VALUE")

  @property
  def MIN_VALUE(self):
    """
    Description:
    ------------
    The MIN_VALUE property returns the smallest positive number possible in JavaScript.

    This static property has a value of 5e-324.

    Usage::

      jsObj.objects.number.get("MyNumber").MIN_VALUE

    Related Pages:

			https://www.w3schools.com/jsref/jsref_min_value.asp
    """
    return JsNumber("Number.MIN_VALUE")

  def isNaN(self):
    """
    Description:
    ------------
    Check whether the value is NaN

    Usage::

      string.parseFloat().isNaN()

    Related Pages:

			https//www.w3schools.com/jsref/jsref_isnan_number.asp

    :return: A Javascript boolean
    """
    from epyk.core.js.primitives import JsBoolean
    return JsBoolean.JsBoolean("Number.isNaN(%s)" % self.varId, isPyData=False)

  def add(self, n):
    """
    Description:
    ------------
    Add a value to a Javascript Number.
    The value will be added and it will return a new number object on the Javascript side

    Usage::

      jsNumber.add(34.5)

    Attributes:
    ----------
    :param n: The number value

    :return: A new Python Javascript Number
    """
    return JsNumber("%s + %s" % (self.varId, n), isPyData=False)

  def sub(self, n):
    """
    Description:
    ------------
    Subtract a value to a Javascript Number.
    The value will be subtracted and it will return a new number object on the Javascript side

    Usage::

      jsNumber.sub(34.5)

    Attributes:
    ----------
    :param n: The number value

    :return: A new Python Javascript Number
    """
    return JsNumber("%s - %s" % (self.varId, n), isPyData=False)

  def toExponential(self):
    """
    Description:
    ------------
    Convert a number into an exponential notation

    Usage::

      jsObj.objects.number.get("MyNumber").toExponential()

    Related Pages:

			https//www.w3schools.com/jsref/jsref_toexponential.asp

    :return: A Javascript Number
    """
    return JsNumber("%s.toExponential()" % self.varId, isPyData=False)

  def toFixed(self, digits=2):
    """
    Description:
    ------------
    Convert a number into a string, keeping only two decimals

    Usage::

      jsObj.objects.number.get("MyNumber").toFixed()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_tofixed.asp

    Attributes:
    ----------
    :param digits: Optional. The number of digits after the decimal point. Default is 2 (2 digits after the decimal point)

    :return: A Javascript Number
    """
    return JsNumber("%s.toFixed(%s)" % (self.varId, digits), isPyData=False)

  def isFinite(self):
    """
    Description:
    ------------
    Check whether a value is a finite number

    Usage::

      jsObj.objects.number.get("MyNumber").isFinite()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_isfinite_number.asp

    :return: A Javascript boolean
    """
    from epyk.core.js.primitives import JsBoolean
    return JsBoolean.JsBoolean("Number.isFinite(%s)" % self.varId, isPyData=False)

  def toPrecision(self, n):
    """
    Description:
    ------------
    Format a number into a specified length:

    Usage::

      rptObj.js.number(varName="myNumber").toPrecision(10) for 5776 returns 5776.000000

    Related Pages:

			https://www.w3schools.com/jsref/jsref_toprecision.asp

    Attributes:
    ----------
    :param n: Optional. The number of digits. If omitted, it returns the entire number (without any formatting)

    :return: A Javascript Number
    """
    return JsNumber("%s.toPrecision(%s)" % (self.varId, n), isPyData=False)

  def __add__(self, value):
    return JsNumber("%s + %s" % (self.varId, value), isPyData=False)

  def __sub__(self, value):
    return JsNumber("%s - %s" % (self.varId, value), isPyData=False)

  def __iadd__(self, value):
    # TODO: Fix this
    self.varData = "%s += %s" % (self.varData, value)
    return self

  def __isub__(self, value):
    # TODO: Fix this
    self.varData = "%s -= %s" % (self.varData, value)
    return self

  def __mul__(self, value):
    return JsNumber("%s * %s" % (self.varId, value), isPyData=False)

  def __truediv__(self, value):
    return JsNumber("%s / %s" % (self.varId, value), isPyData=False)

  def __mod__(self, value):
    return JsNumber("%s %%= %s" % (self.varId, value), isPyData=False)

  def __pow__(self, value):
    return JsNumber("%s = %s" % (self.varId, JsMaths.JsMaths.pow(self, value)))

  @classmethod
  def proto(cls, jsObj, fncName):
    """"""
    jsObj.extendProto(cls, "formatMoney", '''
          var n = this, decPlaces = isNaN(decPlaces = Math.abs(decPlaces)) ? 2 : decPlaces, decSeparator = decSeparator == undefined ? "." : decSeparator,
          thouSeparator = thouSeparator == undefined ? "," : thouSeparator, sign = n < 0 ? "-" : "", i = parseInt(n = Math.abs(+n || 0).toFixed(decPlaces)) + "",
          j = (j = i.length) > 3 ? j % 3 : 0;
          return sign + (j ? i.substr(0, j) + thouSeparator : "") + i.substr(j).replace(/(\d{3})(?=\d)/g, "$1" + thouSeparator) + (decPlaces ? decSeparator + Math.abs(n - i).toFixed(decPlaces).slice(2) : "");
          ''', pmts=["decPlaces", "thouSeparator", "decSeparator"])

  def formatMoney(self, jsObj, decPlaces=0, countryCode='UK'):
    """
    Description:
    ------------
    Wrapper function

    Related Pages:

			https://en.wikipedia.org/wiki/Decimal_separator
    https://docs.oracle.com/cd/E19455-01/806-0169/overview-9/index.html

    Attributes:
    ----------
    :param jsObj: The base Javascript Python object
    :param decPlaces: The number of decimal
    """
    thouSeparator, decSeparator = (",", ".") if countryCode.upper() in ["UK", 'US'] else (" ", ".")
    jsObj.extendProto(self, "formatMoney", '''
      var n = this, decPlaces = isNaN(decPlaces = Math.abs(decPlaces)) ? 2 : decPlaces, decSeparator = decSeparator == undefined ? "." : decSeparator,
      thouSeparator = thouSeparator == undefined ? "," : thouSeparator, sign = n < 0 ? "-" : "", i = parseInt(n = Math.abs(+n || 0).toFixed(decPlaces)) + "",
      j = (j = i.length) > 3 ? j % 3 : 0;
      return sign + (j ? i.substr(0, j) + thouSeparator : "") + i.substr(j).replace(/(\d{3})(?=\d)/g, "$1" + thouSeparator) + (decPlaces ? decSeparator + Math.abs(n - i).toFixed(decPlaces).slice(2) : "");
      ''', pmts=["decPlaces", "thouSeparator", "decSeparator"])
    from epyk.core.js.primitives import JsString
    return JsString.JsString("%s.formatMoney(%s, '%s', '%s')" % (self.varId, decPlaces, thouSeparator, decSeparator), isPyData=False)




from epyk.core import html


class Trees(object):

  def __init__(self, context):
    self.context = context

  def tree(self, data=None, color=None, width=(100, "%"), height=(None, 'px'), htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      data = [{"label": 'test', 'items': [{"label": 'child 1', 'color': 'red'}]}]
      rptObj.ui.lists.tree(data)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTrees.Tree`

    Attributes:
    ----------
    :param data:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_tree = html.HtmlTrees.Tree(self.context.rptObj, data or [], color, width, height, htmlCode, helper, options or {}, profile)
    return html_tree

  def inputs(self, data=None, color=None, width=(100, "%"), height=(None, 'px'),
             htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTrees.TreeInput`

    Attributes:
    ----------
    :param data:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param profile:
    """
    html_tree = html.HtmlTrees.TreeInput(self.context.rptObj, data or [], color, width, height, htmlCode, helper, options or {}, profile)
    return html_tree

  def menu(self, data=None, color=None, width=(100, "%"), height=(None, 'px'), htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlEvent.Menu`

    Attributes:
    ----------
    :param data:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param profile:

    #TODO Ask if this module is still maintained
    """
    html_tree = html.HtmlEvent.Menu(self.context.rptObj, data or [], color, width, height, htmlCode, helper, options or {}, profile)
    return html_tree

  def dropdown(self, recordSet=None, text="", width=(100, "%"), height=(32, 'px'), htmlCode=None,
               helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTrees.DropDown`

    Related Pages:

      http://getbootstrap.com/docs/4.0/components/dropdowns/
      https://www.w3schools.com/bootstrap/tryit.asp?filename=trybs_ref_js_dropdown_multilevel_css&stacked=h
      https://codepen.io/svnt/pen/beEgre

      https://codepen.io/raneio/pen/NbbZEM
      https://www.w3schools.com/howto/tryit.asp?filename=tryhow_css_js_dropdown_hover
      https://codepen.io/antoniputra/pen/BzyWmb

    Attributes:
    ----------
    :param recordSet:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    dftl_options = {"width": 70}
    dftl_options.update(options or {})
    html_d = html.HtmlTrees.DropDown(self.context.rptObj, recordSet, text, width, height, htmlCode, helper,
                                     dftl_options, profile)
    return html_d



"""
CSS Style module for the Search components
"""

from epyk.core.css.styles.classes import CssStyle


class CssSearchExt(CssStyle.Style):
  _attrs = {'box-sizing': 'border-box', 'border-radius': '4px', 'width': '120px',
            'background-repeat': 'no-repeat', 'padding': '5px 20px 5px 40px',
            '-webkit-transition': 'width 0.4s ease-in-out', 'transition': 'width 0.4s ease-in-out'}
  _focus = {'width': '100%', 'outline': 0}

  def customize(self):
    self.css({"background-color": self.rptObj.theme.greys[0]})
    self.hover.css({'color': self.rptObj.theme.greys[-1], 'width': '100%'})


class CssSearch(CssStyle.Style):
  _attrs = {'width': 'auto', 'display': 'inline-block', 'border': 'none', 'background-repeat': 'no-repeat',
            'padding': '5px 20px 5px 40px'}
  _focus = {'outline': 0}

  def customize(self):
    self.css({"background-color": self.rptObj.theme.greys[0], "border-bottom": '1px solid %s' % self.rptObj.theme.greys[3], 'color': self.rptObj.theme.greys[-1]})
    self.hover.css({'color': self.rptObj.theme.greys[-1], 'border-bottom-color': self.rptObj.theme.success[1]})


class CssSearchButton(CssStyle.Style):
  _attrs = {'margin-top': '10px', 'margin-left': '10px', 'display': 'block', 'cursor': 'pointer', 'position': 'absolute'}
  _selectors = {'child': 'i'}

  def customize(self):
    self.css({"color": self.rptObj.theme.greys[5]})



from epyk.core.js.objects import JsNodeDom
from epyk.core.js.primitives import JsObjects
from epyk.core.js.primitives import JsNumber
from epyk.core.js.primitives import JsBoolean
from epyk.core.js.packages import JsQuery
from epyk.core.js.packages import JsQueryUi
from epyk.core.js import JsUtils


class MesuredText(object):
  def __init__(self, varId):
    self.varId = varId

  @property
  def width(self):
    """
#
    """
    return JsNumber.JsNumber("%s.width" % self.varId, isPyData=False)


class RadialGradient(object):
  def __init__(self, varData, varId):
    self.varId, self.varData, self.__set = varId, varData, True
    self._js = []

  def addColorStop(self, stop, color):
    """
    Description:
    ------------
    The addColorStop() method specifies the colors and position in a gradient object.

    The addColorStop() method is used together with createLinearGradient() or createRadialGradient().

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/createRadialGradient

    Attributes:
    ----------
    :param stop: Float. A value between 0.0 and 1.0 that represents the position between start and end in a gradient
    :param color: String. A CSS color value to display at the stop position
    """
    self._js.append("%s.addColorStop(%s, %s)" % (self.varId, stop, JsUtils.jsConvertData(color, None)))
    return self

  def toStr(self):
    if self.__set and self.varData is not None:
      self._js = ["var %s = %s" % (self.varId, self.varData)] + self._js
      self.__set = False
    str_fnc = ";".join(self._js)
    self._js = []
    return str_fnc


class Context2D(object):
  def __init__(self, src):
    self.varId = '%s.getContext("2d")' % src.dom.varId

  def arc(self, x, y, r, sAngle, eAngle, counterclockwise=None):
    """
    Description:
    ------------
    The arc() method creates an arc/curve (used to create circles, or parts of circles).

    Related Pages:

			https://www.w3schools.com/tags/canvas_arc.asp

    Attributes:
    ----------
    :param x: Float. The x-coordinate of the center of the circle
    :param y: Float. The y-coordinate of the center of the circle
    :param r: Float. 	The radius of the circle
    :param sAngle: Float. The starting angle, in radians (0 is at the 3 o'clock position of the arc's circle)
    :param eAngle: Float. The ending angle, in radians
    :param counterclockwise: 	Optional. Specifies whether the drawing should be counterclockwise or clockwise. False is default, and indicates clockwise, while true indicates counter-clockwise.
    """
    return "%s.arc(%s, %s, %s, %s, %s)" % (self.varId, x, y, r, sAngle, eAngle)

  def beginPath(self):
    """
    Description:
    ------------
    The beginPath() method begins a path, or resets the current path.

    Related Pages:

			https://www.w3schools.com/tags/canvas_beginpath.asp
    """
    return "%s.beginPath()" % self.varId

  def bezierCurveTo(self, cp1x, cp1y, cp2x, cp2y, x, y):
    """
    Description:
    ------------
    The bezierCurveTo() method adds a point to the current path by using the specified control points that represent a cubic Bézier curve.

    Related Pages:

			https://www.w3schools.com/tags/canvas_beziercurveto.asp

    Attributes:
    ----------
    :param cp1x: Float. The x-coordinate of the first Bézier control point
    :param cp1y: Float. The y-coordinate of the first Bézier control point
    :param cp2x: Float. The x-coordinate of the second Bézier control point
    :param cp2y: Float. The y-coordinate of the second Bézier control point
    :param x: Float. The x-coordinate of the ending point
    :param y: Float. The y-coordinate of the ending point
    """
    return "%s.bezierCurveTo(%s, %s, %s, %s, %s, %s)" % (self.varId, cp1x, cp1y, cp2x, cp2y, x, y)

  def clearRect(self, x, y, width, height):
    """
    Description:
    ------------
    The clearRect() method clears the specified pixels within a given rectangle.

    Related Pages:

			https://www.w3schools.com/tags/canvas_clearrect.asp

    Attributes:
    ----------
    :param x: Float. The x-coordinate of the upper-left corner of the rectangle to clear
    :param y: Float. The y-coordinate of the upper-left corner of the rectangle to clear
    :param width: Float. The width of the rectangle to clear, in pixels
    :param height: Float. The height of the rectangle to clear, in pixels
    """
    return "%s.clearRect(%s, %s, %s, %s)" % (self.varId, x, y, width, height)

  def clip(self):
    """
    Description:
    ------------
    The clip() method clips a region of any shape and size from the original canvas.

    Related Pages:

			https://www.w3schools.com/tags/canvas_clip.asp
    """
    return "%s.clip()" % self.varId

  def closePath(self):
    """
    Description:
    ------------
    The closePath() method creates a path from the current point back to the starting point.

    Related Pages:

			https://www.w3schools.com/tags/canvas_closepath.asp
    """
    return "%s.closePath()" % self.varId

  def createPattern(self, image, repeatType):
    """
    Description:
    ------------
    The createPattern() method repeats the specified element in the specified direction.

    The element can be an image, video, or another <canvas> element.

    Related Pages:

			https://www.w3schools.com/tags/canvas_createpattern.asp

    Attributes:
    ----------
    :param image: Specifies the image, canvas, or video element of the pattern to use
    :param repeatType: Default. The pattern repeats both horizontally and vertically
    """
    image = JsUtils.jsConvertData(image, None)
    repeatType = JsUtils.jsConvertData(repeatType, None)
    return "%s.createPattern(%s, %s)" % (self.varId, image, repeatType)

  def createRadialGradient(self, x0, y0, r0, x1, y1, r1, varName):
    """
    Description:
    ------------
    The createRadialGradient() method is specified by six parameters, three defining the gradient's start circle, and three defining the end circle.

    Related Pages:

			https://www.w3schools.com/tags/canvas_createradialgradient.asp

    Attributes:
    ----------
    :param x0: Float. The x-coordinate of the starting circle of the gradient
    :param y0: Float. 	The y-coordinate of the starting circle of the gradient
    :param r0: Float. The radius of the starting circle
    :param x1: Float. The x-coordinate of the ending circle of the gradient
    :param y1: Float. The y-coordinate of the ending circle of the gradient
    :param r1: Float. The radius of the ending circle
    :param varName: String. The object reference on the Javascript side
    """
    gradient_id = "%s.createRadialGradient(%s, %s, %s, %s, %s, %s)" % (self.varId, x0, y0, r0, x1, y1, r1)
    return RadialGradient(varData=gradient_id, varId=varName)

  def createLinearGradient(self, x0, y0, x1, y1, varName):
    """
    Description:
    ------------
    The createLinearGradient() method creates a linear gradient object.

    The gradient can be used to fill rectangles, circles, lines, text, etc.

    Usage::

      canvas.ctx.createLinearGradient(0, 0, canvas.dom.width, 0, "test").
    addColorStop("0", "magenta").addColorStop("0.5", "blue").addColorStop("1.0", "red"),
    canvas.ctx.strokeStyle(rptObj.js.object("test")),
    canvas.ctx.strokeText("Hello World!", 10, 50),

    Related Pages:

			https://www.w3schools.com/tags/canvas_createlineargradient.asp

    Attributes:
    ----------
    :param x0: Float. The x-coordinate of the start point of the gradient
    :param y0: Float. The y-coordinate of the start point of the gradient
    :param x1: Float. The x-coordinate of the end point of the gradient
    :param y1: Float. The y-coordinate of the end point of the gradient
    :param varName: String. The object reference on the Javascript side
    """
    gradient_id = "%s.createLinearGradient(%s, %s, %s, %s)" % (self.varId, x0, y0, x1, y1)
    return RadialGradient(varData=gradient_id, varId=varName)

  def fill(self, color=None):
    """
    Description:
    ------------
    The fill() method fills the current drawing (path). The default color is black.

    Related Pages:

			https://www.w3schools.com/tags/canvas_fill.asp

    Attributes:
    ----------
    :param color: Optional. String the color to be added to the fillStyle
    """
    if color is None:
      return "%s.fill()" % self.varId

    return "%s; %s.fill()" % (self.fillStyle(color), self.varId)

  def fillStyle(self, color):
    """
    Description:
    ------------
    The fillStyle property sets or returns the color, gradient, or pattern used to fill the drawing.

    Related Pages:

			https://www.w3schools.com/tags/canvas_fillstyle.asp

    Attributes:
    ----------
    :param color: String. The colur definition
    """
    color = JsUtils.jsConvertData(color, None)
    return "%s.fillStyle = %s" % (self.varId, color)

  def fillText(self, text, x, y, maxWidth=None, fillStyle=None):
    """
    Description:
    ------------
    The fillText() method draws filled text on the canvas. The default color of the text is black.

    Related Pages:

			https://www.w3schools.com/tags/canvas_filltext.asp

    Attributes:
    ----------
    :param text: String. Specifies the text that will be written on the canvas
    :param x: Float. The x coordinate where to start painting the text (relative to the canvas)
    :param y: Float. The y coordinate where to start painting the text (relative to the canvas)
    :param maxWidth: Optional. The maximum allowed width of the text, in pixels
    """
    if isinstance(text, list):
      text = " + ".join([str(JsUtils.jsConvertData(t, None)) for t in text])
    else:
      text = JsUtils.jsConvertData(text, None)
    if fillStyle is not None:
      return "%s; %s.fillText(%s, %s, %s)" % (self.fillStyle(fillStyle), self.varId, text, x, y)

    if maxWidth is not None:
      return "%s.fillText(%s, %s, %s, %s)" % (self.varId, text, x, y, maxWidth)

    return "%s.fillText(%s, %s, %s)" % (self.varId, text, x, y)

  def lineCap(self, value):
    """
    Description:
    ------------
    The lineCap property sets or returns the style of the end caps for a line.

    Related Pages:

			https://www.w3schools.com/tags/canvas_linecap.asp

    Attributes:
    ----------
    :param value:
    """
    value = JsUtils.jsConvertData(value, None)
    return "%s.lineCap = %s" % (self.varId, value)

  def lineJoin(self, value):
    """
    Description:
    ------------
    The lineJoin property sets or returns the type of corner created, when two lines meet.

    Related Pages:

			https://www.w3schools.com/tags/canvas_linejoin.asp

    Attributes:
    ----------
    :param value:
    """
    value = JsUtils.jsConvertData(value, None)
    return "%s.lineJoin = %s" % (self.varId, value)

  def lineTo(self, x, y):
    """
    Description:
    ------------
    Adds a new point and creates a line to that point from the last specified point in the canvas

    Related Pages:

			https://www.w3schools.com/tags/canvas_lineto.asp

    Attributes:
    ----------
    :param x: Float. The x-coordinate of where to create the line to
    :param y: Float. The y-coordinate of where to create the line to
    """
    return "%s.lineTo(%s, %s)" % (self.varId, x, y)

  def lineWidth(self, value):
    """
    Description:
    ------------
    The lineWidth property sets or returns the current line width, in pixels.

    Related Pages:

			https://www.w3schools.com/tags/canvas_linewidth.asp

    Attributes:
    ----------
    :param value: Float. The current line width, in pixels
    """
    return "%s.lineWidth = %s" % (self.varId, value)

  def moveTo(self, x, y):
    """
    Description:
    ------------
    Moves the path to the specified point in the canvas, without creating a line

    Related Pages:

			https://www.w3schools.com/tags/canvas_moveto.asp

    Attributes:
    ----------
    :param x: Float. The x-coordinate of where to move the path to
    :param y: Float. The y-coordinate of where to move the path to
    """
    return "%s.moveTo(%s, %s)" % (self.varId, x, y)

  def rect(self, x, y, width, height):
    """
    Description:
    ------------
    The rect() method creates a rectangle.

    Related Pages:

			https://www.w3schools.com/tags/canvas_rect.asp

    Attributes:
    ----------
    :param x: Float. The x-coordinate of the upper-left corner of the rectangle
    :param y: Float. The y-coordinate of the upper-left corner of the rectangle
    :param width: Float. The width of the rectangle, in pixels
    :param height: Float. 	The height of the rectangle, in pixels
    """
    return "%s.rect(%s, %s, %s, %s)" % (self.varId, x, y, width, height)

  def fillRect(self, x, y, width, height):
    """
    Description:
    ------------
    The rect() method creates a rectangle.

    Related Pages:

			https://www.w3schools.com/tags/canvas_rect.asp

    Attributes:
    ----------
    :param x: Float. The x-coordinate of the upper-left corner of the rectangle
    :param y: Float. The y-coordinate of the upper-left corner of the rectangle
    :param width: Float. The width of the rectangle, in pixels
    :param height: Float. 	The height of the rectangle, in pixels
    """
    return "%s.fillRect(%s, %s, %s, %s)" % (self.varId, x, y, width, height)

  def font(self, fontStyle):
    """
    Description:
    ------------
    The font property sets or returns the current font properties for text content on the canvas.

    Related Pages:

			https://www.w3schools.com/tags/canvas_font.asp

    Attributes:
    ----------
    :param fontStyle: String. The font properties
    """
    fontStyle = JsUtils.jsConvertData(fontStyle, None)
    return "%s.font = %s" % (self.varId, fontStyle)

  def globalAlpha(self, number):
    """
    Description:
    ------------
    The globalAlpha property sets or returns the current alpha or transparency value of the drawing.

    The globalAlpha property value must be a number between 0.0 (fully transparent) and 1.0 (no transparancy).

    Related Pages:

			https://www.w3schools.com/tags/canvas_globalalpha.asp

    Attributes:
    ----------
    :param number: The transparency value. Must be a number between 0.0 (fully transparent) and 1.0 (no transparancy)
    """
    return "%s.globalAlpha = %s" % (self.varId, number)

  def isPointInPath(self, x, y):
    """
    Description:
    ------------
    The isPointInPath() method returns true if the specified point is in the current path, otherwise false.

    Related Pages:

			https://www.w3schools.com/tags/canvas_ispointinpath.asp

    Attributes:
    ----------
    :param x: Float. The x-coordinate to test
    :param y: Float. The y-coordinate to test
    """
    return JsBoolean.JsBoolean("%s.isPointInPath(%s, %s)" % (self.varId, x, y), isPyData=False)

  def measureText(self, text):
    """
    Description:
    ------------
    The measureText() method returns an object that contains the width of the specified text, in pixels.

    Related Pages:

			https://www.w3schools.com/tags/canvas_measuretext.asp

    Attributes:
    ----------
    :param text: String. The text
    """
    if isinstance(text, list):
      text = " + ".join([str(JsUtils.jsConvertData(t, None)) for t in text])
    else:
      text = JsUtils.jsConvertData(text, None)
    return MesuredText("%s.measureText(%s)" % (self.varId, text))

  def rotate(self, angle):
    """
    Description:
    ------------
    The rotate() method rotates the current drawing.

    Related Pages:

			https://www.w3schools.com/tags/canvas_rotate.asp

    Attributes:
    ----------
    :param angle: Float. The angle number
    """
    return "%s.rotate(%s)" % (self.varId, angle)

  def scale(self, scalewidth, scaleheight):
    """
    Description:
    ------------
    The scale() method scales the current drawing, bigger or smaller.

    Related Pages:

			https://www.w3schools.com/tags/canvas_scale.asp

    Attributes:
    ----------
    :param scalewidth: Float. Scales the width of the current drawing (1=100%, 0.5=50%, 2=200%, etc.)
    :param scaleheight: Float. Scales the height of the current drawing (1=100%, 0.5=50%, 2=200%, etc.)
    """
    return "%s.scale(%s, %s)" % (self.varId, scalewidth, scaleheight)

  def shadowBlur(self, value):
    """
    Description:
    ------------
    The shadowBlur property sets or returns the blur level for shadows.

    Related Pages:

			https://www.w3schools.com/tags/canvas_shadowblur.asp

    Attributes:
    ----------
    :param value: Float. The blur level for the shadow
    """
    return "%s.shadowBlur = %s" % (self.varId, value)

  def shadowColor(self, color):
    """
    Description:
    ------------
    The shadowColor property sets or returns the color to use for shadows.

    Related Pages:

			https://www.w3schools.com/tags/canvas_shadowcolor.asp

    Attributes:
    ----------
    :param color: String. the color code
    """
    color = JsUtils.jsConvertData(color, None)
    return "%s.shadowColor = %s" % (self.varId, color)

  def shadowOffsetX(self, value):
    """
    Description:
    ------------
    The shadowOffsetX property sets or returns the horizontal distance of the shadow from the shape.

    Related Pages:

			https://www.w3schools.com/tags/canvas_shadowoffsetx.asp

    Attributes:
    ----------
    :param value: A positive or negative number that defines the horizontal distance of the shadow from the shape
    """
    return "%s.shadowOffsetX = %s" % (self.varId, value)

  def shadowOffsetY(self, value):
    """
    Description:
    ------------
    The shadowOffsetY property sets or returns the vertical distance of the shadow from the shape.

    Related Pages:

			https://www.w3schools.com/tags/canvas_shadowoffsety.asp

    Attributes:
    ----------
    :param value: A positive or negative number that defines the vertical distance of the shadow from the shape
    """
    return "%s.shadowOffsetY = %s" % (self.varId, value)

  def stroke(self):
    """
    Description:
    ------------
    The stroke() method actually draws the path you have defined with all those moveTo() and lineTo() methods.
    The default color is black.
    """
    return "%s.stroke()" % self.varId

  def strokeWeight(self, val):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param val:
    """
    return "%s.strokeWeight = %s" % (self.varId, val)

  def strokeStyle(self, color="#000000"):
    """
    Description:
    ------------
    The strokeStyle property sets or returns the color, gradient, or pattern used for strokes.

    Related Pages:

			https://www.w3schools.com/tags/canvas_strokestyle.asp

    Attributes:
    ----------
    :param color: A CSS color value that indicates the stroke color of the drawing. Default value is #000000
    """
    return "%s.strokeStyle = %s" % (self.varId, JsUtils.jsConvertData(color, None))

  def strokeRect(self):
    """
    Description:
    ------------
    The strokeRect() method draws a rectangle (no fill). The default color of the stroke is black.

    Related Pages:

			https://www.w3schools.com/tags/canvas_strokestyle.asp
    """

  def strokeText(self, text, x, y, maxWidth=None):
    """
    Description:
    ------------
    The strokeText() method draws text (with no fill) on the canvas. The default color of the text is black.

    Related Pages:

			https://www.w3schools.com/tags/canvas_stroketext.asp

    Attributes:
    ----------
    :param text: String. Specifies the text that will be written on the canvas
    :param x: Float. The x coordinate where to start painting the text (relative to the canvas)
    :param y: Float. The y coordinate where to start painting the text (relative to the canvas)
    :param maxWidth: Optional. The maximum allowed width of the text, in pixels
    """
    if isinstance(text, list):
      text = " + ".join([str(JsUtils.jsConvertData(t, None)) for t in text])
    else:
      text = JsUtils.jsConvertData(text, None)
    return "%s.strokeText(%s, %s, %s)" % (self.varId, text, x, y)

  def textAlign(self, position):
    """
    Description:
    ------------
    The textAlign property sets or returns the current alignment for text content, according to the anchor point.

    Related Pages:

			https://www.w3schools.com/tags/canvas_textalign.asp

    Attributes:
    ----------
    :param position: String. The context align
    """
    return "%s.textAlign = %s" % (self.varId, JsUtils.jsConvertData(position, None))

  def translate(self, x, y):
    """
    Description:
    ------------
    The translate() method remaps the (0,0) position on the canvas.

    Related Pages:

			https://www.w3schools.com/tags/canvas_translate.asp

    Attributes:
    ----------
    :param x: Float. The value to add to horizontal (x) coordinates
    :param y: Float. The value to add to vertical (y) coordinates
    """
    return "%s.translate(%s, %s)" % (self.varId, x, y)

  def textBaseline(self, position):
    """
    Description:
    ------------
    The textBaseline property sets or returns the current text baseline used when drawing text.

    Related Pages:

			https://www.w3schools.com/tags/canvas_textbaseline.asp

    Attributes:
    ----------
    :param position: String. The context baseline
    """
    return "%s.textBaseline = %s" % (self.varId, JsUtils.jsConvertData(position, None))

  def drawImage(self, img, x, y, width, height):
    """
    Description:
    ------------
    The drawImage() method draws an image, canvas, or video onto the canvas.

    Related Pages:

			https://www.w3schools.com/tags/canvas_drawimage.asp

    Attributes:
    ----------
    :param img: String. Specifies the image, canvas, or video element to use
    :param x: Float. Optional. The x coordinate where to place the image on the canvas
    :param y: Float. Optional. The y coordinate where to place the image on the canvas
    :param width: Float. Optional. The width of the image to use (stretch or reduce the image)
    :param height: Float. Optional. The height of the image to use (stretch or reduce the image)
    """
    pass

  def getImageData(self, x, y, width, height):
    """
    Description:
    ------------
    The getImageData() method returns an ImageData object that copies the pixel data for the specified rectangle on a canvas.

    Related Pages:

			https://www.w3schools.com/tags/canvas_getimagedata.asp

    Attributes:
    ----------
    :param x: Float. The x coordinate (in pixels) of the upper-left corner to start copy from
    :param y: Float. The y coordinate (in pixels) of the upper-left corner to start copy from
    :param width: Float. The width of the rectangular area you will copy
    :param height: Float. The height of the rectangular area you will copy
    """
    return "%s.getImageData(%s, %s, %s, %s)" % (self.varId, x, y, width, height)


class Canvas(JsNodeDom.JsDoms):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.htmlCode = varName if varName is not None else htmlObj.htmlCode
    self.varName, self.varData, self.__var_def = "document.getElementById('%s')" % self.htmlCode, "", None
    self._src, self._report = htmlObj, report
    self._js, self.__2d_context = [], None
    self._jquery, self._jquery_ui = None, None

  @property
  def width(self):
    """

    :return:
    """
    return JsNumber.JsNumber("%s.width" % self.varId)

  @property
  def getContext2D(self):
    """
    Description:
    ------------
    The HTMLCanvasElement.getContext() method returns a drawing context on the canvas, or null if the context identifier is not supported.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/HTMLCanvasElement/getContext

    :rtype: Context2D
    """
    if self.__2d_context is None:
      self.__2d_context = Context2D(self._src)
    return self.__2d_context

  @property
  def val(self):
    """
    Description:
    ------------

    :return:
    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s.value, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
      self.htmlCode, self.varName))

  @property
  def events(self):
    """
    Description:
    ------------

    :rtype: JsNodeDom.JsDomEvents
    """
    return JsNodeDom.JsDomEvents(self._src)

  @property
  def jquery(self):
    """
    Description:
    ------------

    :rtype: JsQuery.JQuery
    """
    if self._jquery is None:
      self._jquery = JsQuery.JQuery(src=self._src, selector=JsQuery.decorate_var("#%s" % self._src.htmlCode))
    return self._jquery

  @property
  def jquery_ui(self):
    """

    :return:
    :rtype: JsQuery.JQuery
    """
    if self._jquery_ui is None:
      self._jquery_ui = JsQueryUi.JQueryUI(self._src, selector=JsQuery.decorate_var("#%s" % self._src.htmlCode))
    return self._jquery_ui




from epyk.core.js import JsUtils
from epyk.core.js.packages import JsPackage


class JsChartDCLinks(object):
  def __init__(self, data, js_src, data_schema=None, profile=False):
    self._js_src, self._data_schema, self.profile, self._data = js_src, data_schema, profile, data

  def __register_records_fnc(self, fnc_name, fnc_def, fnc_pmts=None, profile=False):
    """
    This function will attach to the report object only the javascript functions used during the report

    :param fnc_name: A String with the Javascript function name to be defined
    :param fnc_def: A String with the Javascript function content
    :param fnc_pmts: A list of parameters
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    fnc_pmts = ["data"] + (fnc_pmts or [])
    if not fnc_name in self._js_src.get('js', {}).get('functions', {}):
      if profile:
        content = "var result = []; %s;return result" % JsUtils.cleanFncs(fnc_def)
      else:
        content = "var result = []; %s;return result" % JsUtils.cleanFncs(fnc_def)
      self._js_src.setdefault('js', {}).setdefault('functions', {})[fnc_name] = {'content': content, 'pmt': fnc_pmts}

  def line(self):
    pass


class JsBase(JsPackage):
  """
  Base Class for the DC Charts
  """
  def version(self, no=None):
    """
    Return or change the underlying Javascript package version

    :param no: Optional. The new package version to force

    :return: The package version used
    """
    if no is not None:
      self

    return

  def addFilterHandler(self, addFilterHandler):
    """
    Set or get the add filter handler. The add filter handler is a function that adds a filter to the chart's filter list.
    Using a custom add filter handler allows you to change the way filters are added or perform additional work when adding a filter, e.g.
    when using a filter server other than crossfilter.

    :param addFilterHandler:

    :return:
    """
    self._js.append("addFilterHandler(%s)" % addFilterHandler)
    return self

  def chartGroup(self, groupId=None):
    """
    Get or set the chart group to which this chart belongs.
    Chart groups are rendered or redrawn together since it is expected they share the same underlying crossfilter data set.

    Related Pages:

			https://dc-js.github.io/dc.js/docs/html/dc.baseMixin.html

    :param groupId: Optional. The group ID

    :return:
    """
    return

  def data(self, callback):
    """
    Set the data callback or retrieve the chart's data set.
    The data callback is passed the chart's group and by default will return group.all. This behavior may be modified to, for instance, return only the top 5 groups.

    :param callback: Optinal

    :return:
    """

  def chartID(self):
    """
    Returns the internal numeric ID of the chart.

    Related Pages:

			https://dc-js.github.io/dc.js/docs/html/dc.baseMixin.html

    :return:
    """

  def x(self, xScale):
    """

    :return:
    """
    self._js.append("x(%s)" % xScale)
    return self

  def y(self, yScale):
    """
    Get or set the y scale. The y scale is typically automatically determined by the chart implementation.

    :return:
    """
    self._js.append("y(%s)" % yScale)
    return self

  def yAxis(self, yAxis=None):
    """
    Set or get the y axis used by the coordinate grid chart instance. This function is most useful when y axis customization is required.
    Depending on useRightYAxis the y axis in dc.js is an instance of either d3.axisLeft or d3.axisRight; therefore it supports any valid d3 axis manipulation.

    :return:
    """
    self._js.append("yAxis(%s)" % yAxis)
    return self

  def xUnits(self, unit=None):
    """
    Set or get the xUnits function.
    The coordinate grid chart uses the xUnits function to calculate the number of data projections on the x axis such as the number of bars for a bar chart or the number of dots for a line chart.

    :return:
    """
    self._js.append("xUnits(%s)" % unit)
    return self

  def brushOn(self, brushOn=True):
    """
    urn on/off the brush-based range filter.
    When brushing is on then user can drag the mouse across a chart with a quantitative scale to perform range filtering based on the extent of the brush, or click on the bars of an ordinal bar chart or slices of a pie chart to filter and un-filter them.

    Related Pages:

			https://dc-js.github.io/dc.js/docs/html/dc.coordinateGridMixin.html

    :param brushOn:
    :return:
    """
    self._js.append("brushOn(%s)")
    return self

  def dimension(self, dimension):
    """
    Set or get the dimension attribute of a chart. In dc, a dimension can be any valid crossfilter dimension

    :param: A Cross filter dimension
    :return:
    """

  def expireCache(self):
    """
    Expire the internal chart cache. dc charts cache some data internally on a per chart basis to speed up rendering and avoid unnecessary calculation

    :return:
    """

  def filter(self, filter):
    """
    Filter the chart by the given parameter, or return the current filter if no input parameter is given.

    :param filter: Filter can be an Array, a single value or a DC filter object

    :return:
    """

  def filterAll(self):
    """
    Clear all filters associated with this chart. The same effect can be achieved by calling chart.filter(null).

    :return:
    """

  def filterHandler(self, filterHandler):
    """
    Set or get the filter handler.
    The filter handler is a function that performs the filter action on a specific dimension.
    Using a custom filter handler allows you to perform additional logic before or after filtering.

    :return:
    """

  def filterPrinter(self, filterPrinter):
    """
    Set or get the filter printer function.
    The filter printer function is used to generate human friendly text for filter value(s) associated with the chart instance.
    The text will get shown in the `.filter element; see turnOnControls.

    :param filterPrinter:
    :return:
    """

  def filters(self):
    """
    Returns all current filters.
    This method does not perform defensive cloning of the internal filter array before returning, therefore any modification of the returned array will effect the chart's internal filter storage.

    :return:
    """

  def group(self):
    """
    Set or get the group attribute of a chart.
    In dc a group is a crossfilter group.
    Usually the group should be created from the particular dimension associated with the same chart. If a value is given, then it will be used as the new group.

    :return:
    """

  def hasFilter(self, filter=None):
    """
    Check whether any active filter or a specific filter is associated with particular chart instance. This function is not

    :param filter:
    :return:
    """

  def height(self, height=None):
    """
    Set or get the height attribute of a chart.
    The height is applied to the SVGElement generated by the chart when rendered (or re-rendered).
    If a value is given, then it will be used to calculate the new height and the chart returned for method chaining.
    The value can either be a numeric, a function, or falsy.

    :param height:

    :return:
    """
    self._js.append("height(%s)" % height)
    return self

  def width(self, width=None):
    """

    :param width:
    :return:
    """
    self._js.append("width(%s)" % width)
    return self

  def label(self):
    """
    Set or get the label function. The chart class will use this function to render labels for each child element in the chart, e.g. slices in a pie chart or bubbles in a bubble chart.
    Not every chart supports the label function, for example line chart does not use this function at all. By default, enables labels; pass false for the second parameter if this is not desired.

    :return:
    """

  def render(self):
    """
    Invoking this method will force the chart to re-render everything from scratch.
    Generally it should only be used to render the chart for the first time on the page or if you want to make sure everything is redrawn from scratch instead of relying on the default incremental redrawing behaviour.

    :return:
    """
    self._js.append("render()")
    return self


class JsLine(JsBase):
  """
  Configuration for a Lines Chart in DC
  """
  alias = 'line'
  name = 'Bars'
  jsCls = 'lineChart'
  _attrs = {}


class JsBar(JsBase):
  """
  Configuration for a Bars Chart in DC
  """
  alias = 'bar'
  name = 'Bars'
  jsCls = 'barChart'
  _attrs = {}

  def xAxisLabel(self, labelText=None, padding=None):
    """
    Set or get the x axis label. If setting the label, you may optionally include additional padding to the margin to make room for the label.
    By default the padded is set to 12 to accomodate the text height.

    :param labelText:
    :param padding:

    :return:
    """
    self._js.append("xAxisLabel(%s)" % labelText)
    return self

  def yAxisLabel(self, labelText=None, padding=None):
    """
    Set or get the x axis label. If setting the label, you may optionally include additional padding to the margin to make room for the label.
    By default the padded is set to 12 to accomodate the text height.

    :param labelText:
    :param padding:

    :return:
    """
    self._js.append("xAxisLabel(%s)" % labelText)
    return self


class JsPie(JsBase):
  """
  Configuration for a Pies Chart in DC
  """
  alias = 'pie'
  name = 'Pie Chart'
  jsCls = 'pieChart'
  _attrs = {}

  def x(self):
    raise Exception("Does not exist for this chart")


  def emptyTitle(self, title):
    """
    Title to use for the only slice when there is no data.

    Related Pages:

			https://dc-js.github.io/dc.js/docs/html/dc.pieChart.html

    :param title:

    :return:
    """
    return


class JsBuble(JsBase):
  """
  Configuration for a Pies Chart in DC
  """
  alias = 'bubble'
  name = 'Bubble Chart'
  jsCls = 'bubbleChart'
  _attrs = {}

  def x(self):
    raise Exception("Does not exist for this chart")


class JsRow(JsBase):
  """
  Configuration for a Pies Chart in DC
  """
  alias = 'row'
  name = 'Horizontal Bar Chart'
  jsCls = 'rowChart'
  _attrs = {}





from epyk.core.html import tables as html_tables


class AgGrid(object):

  def __init__(self, context):
    self.parent = context

  def table(self, records=None, cols=None, rows=None, width=(100, '%'), height=(300, 'px'), htmlCode=None, options=None, profile=None):
    """

    :param records:
    :param cols:
    :param rows:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param profile:
    """
    cols = cols or []
    rows = rows or []
    if not cols and not rows:
      cols = list(records[0].keys())

    table_options_dflts = {'headerHeight': 30, 'rowHeight': '50'}
    if options is not None:
      table_options_dflts.update(options)

    table = html_tables.HtmlTableAgGrid.Table(self.parent.context.rptObj, records, width, height, htmlCode, table_options_dflts, profile)
    for c in cols + rows:
      table.add_column(c)
    return table




from epyk.core.js.packages import JsPackage
from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects


class TravelMode():
  pass


class DirectionsStatus():
  pass


class DirectionsService():
  pass


class DirectionsRenderer():
  pass


class Marker():

  @property
  def id(self):
    """
    """
    return self._config_get(10)

  @id.setter
  def id(self, num):
    self._config(num)

  @property
  def html(self):
    """
    """
    return self._config_get(10)

  @html.setter
  def html(self, num):
    self._config(num)


class GoogleMapsAPI(JsPackage):
  lib_alias = {'js': 'google-maps'}
  lib_set_var = False

  def marker(self, latitude, longitude):
    """

    :param latitude:
    :param longitude:
    """
    Marker({"lat": latitude, 'lng': longitude})
    return

  def setMapTypeId(self, mapTypeId):
    """

    https://developers.google.com/maps/documentation/javascript/maptypes

    :param mapTypeId:
    """
    mapTypeId = JsUtils.jsConvertData(mapTypeId, None)
    return JsObjects.JsVoid("%s.setMapTypeId(%s)" % (self.varId, mapTypeId))

  def setTilt(self, value):
    """
    https://developers.google.com/maps/documentation/javascript/maptypes

    :param value:
    :return:
    """
    return JsObjects.JsVoid("%s.setTilt(%s)" % (self.varId, value))

  def getTile(self, value):
    """
    is called whenever the API determines that the map needs to display new tiles for the given viewport.
    The getTile() method must have the following signature:

    https://developers.google.com/maps/documentation/javascript/maptypes

    :param value:
    :return:
    """
    return JsObjects.JsVoid("%s.getTile(%s)" % (self.varId, value))

  def setHeading(self, value):
    """
    https://developers.google.com/maps/documentation/javascript/maptypes

    :param value:
    :return:
    """
    return JsObjects.JsVoid("%s.setHeading(%s)" % (self.varId, value))

  def getHeading(self):
    """

    https://developers.google.com/maps/documentation/javascript/maptypes

    :return:
    """
    return JsObjects.JsNumber.JsNumber("%s.getHeading()" % self.varId)



from epyk.core.py import Pyk
from epyk.core.Page import Report
pyk_rpt = Pyk.requires(r'C:\Users\nelso\PycharmProjects\epyk-ui\epyk\tests\pyk_report.py')


rpt_obj = Report()

rpt_obj.ui.text("Here we have a text followed by a normal button")

rpt_obj.register(pyk_rpt.button)

rpt_obj.ui.text("Then we have an import button")

rpt_obj.register(pyk_rpt.imp)

print(rpt_obj.outs.html_file())



"""
Related Pages:

		https://www.w3schools.com/charsets/ref_utf_math.asp
"""

FOR_ALL = '&#8704'
COMPLEMENT = '&#8705'
PARTIAL_DIFFERENTIAL = '&#8706'
THERE_EXISTS = '&#8707'
THERE_DOES_NOT_EXIST = '&#8708'
EMPTY_SET = '&#8709'
INCREMENT = '&#8710'
NABLA = '&#8711'
ELEMENT_OF = '&#8712'
NOT_AN_ELEMENT_OF = '&#8713'
SMALL_ELEMENT_OF = '&#8714'
CONTAINS_AS_MEMBER = '&#8715'
DOES_NOT_CONTAIN_AS_MEMBER = '&#8716'
SMALL_CONTAINS_AS_MEMBER = '&#8717'
END_OF_PROOF = '&#8718'
N_ARY_PRODUCT = '&#8719'
N_ARY_COPRODUCT = '&#8720'
N_ARY_SUMMATION = '&#8721'
MINUS_SIGN = '&#8722'
MINUS_OR_PLUS_SIGN = '&#8723'
DOT_PLUS = '&#8724'
DIVISION_SLASH = '&#8725'
SET_MINUS = '&#8726'
ASTERISK_OPERATOR = '&#8727'
RING_OPERATOR = '&#8728'
BULLET_OPERATOR = '&#8729'
SQUARE_ROOT = '&#8730'
CUBE_ROOT = '&#8731'
FOURTH_ROOT = '&#8732'
PROPORTIONAL_TO = '&#8733'
INFINITY = '&#8734'
RIGHT_ANGLE = '&#8735'
ANGLE = '&#8736'
MEASURED_ANGLE = '&#8737'
SPHERICAL_ANGLE = '&#8738'
DIVIDES = '&#8739'
DOES_NOT_DIVIDE = '&#8740'
PARALLEL_TO = '&#8741'
NOT_PARALLEL_TO = '&#8742'
LOGICAL_AND = '&#8743'
LOGICAL_OR = '&#8744'
INTERSECTION = '&#8745'
UNION = '&#8746'
INTEGRAL = '&#8747'
DOUBLE_INTEGRAL = '&#8748'
TRIPLE_INTEGRAL = '&#8749'
CONTOUR_INTEGRAL = '&#8750'
SURFACE_INTEGRAL = '&#8751'
VOLUME_INTEGRAL = '&#8752'
CLOCKWISE_INTEGRAL = '&#8753'
CLOCKWISE_CONTOUR_INTEGRAL = '&#8754'
ANTICLOCKWISE_CONTOUR_INTEGRAL = '&#8755'
THEREFORE = '&#8756'
BECAUSE = '&#8757'
RATIO = '&#8758'
PROPORTION = '&#8759'
DOT_MINUS = '&#8760'
EXCESS = '&#8761'
GEOMETRIC_PROPORTION = '&#8762'
HOMOTHETIC = '&#8763'
TILDE_OPERATOR = '&#8764'
REVERSED_TILDE = '&#8765'
INVERTED_LAZY_S = '&#8766'
SINE_WAVE = '&#8767'
WREATH_PRODUCT = '&#8768'
NOT_TILDE = '&#8769'
MINUS_TILDE = '&#8770'
ASYMPTOTICALLY_EQUAL_TO = '&#8771'
NOT_ASYMPTOTICALLY_EQUAL_TO = '&#8772'
APPROXIMATELY_EQUAL_TO = '&#8773'
APPROXIMATELY_BUT_NOT_ACTUALLY_EQUAL_TO = '&#8774'
NEITHER_APPROXIMATELY_NOR_ACTUALLY_EQUAL_TO = '&#8775'
ALMOST_EQUAL_TO = '&#8776'
NOT_ALMOST_EQUAL_TO = '&#8777'
ALMOST_EQUAL_OR_EQUAL_TO = '&#8778'
TRIPLE_TILDE = '&#8779'
ALL_EQUAL_TO = '&#8780'
EQUIVALENT_TO = '&#8781'
GEOMETRICALLY_EQUIVALENT_TO = '&#8782'
DIFFERENCE_BETWEEN = '&#8783'
APPROACHES_THE_LIMIT = '&#8784'
GEOMETRICALLY_EQUAL_TO = '&#8785'
APPROXIMATELY_EQUAL_TO_OR_THE_IMAGE_OF = '&#8786'
IMAGE_OF_OR_APPROXIMATELY_EQUAL_TO = '&#8787'
COLON_EQUALS = '&#8788'
EQUALS_COLON = '&#8789'
RING_IN_EQUAL_TO = '&#8790'
RING_EQUAL_TO = '&#8791'
CORRESPONDS_TO = '&#8792'
ESTIMATES = '&#8793'
EQUIANGULAR_TO = '&#8794'
STAR_EQUALS = '&#8795'
DELTA_EQUAL_TO = '&#8796'
EQUAL_TO_BY_DEFINITION = '&#8797'
MEASURED_BY = '&#8798'
QUESTIONED_EQUAL_TO = '&#8799'
NOT_EQUAL_TO = '&#8800'
IDENTICAL_TO = '&#8801'
NOT_IDENTICAL_TO = '&#8802'
STRICTLY_EQUIVALENT_TO = '&#8803'
LESS_THAN_OR_EQUAL_TO = '&#8804'
GREATER_THAN_OR_EQUAL_TO = '&#8805'
LESS_THAN_OVER_EQUAL_TO = '&#8806'
GREATER_THAN_OVER_EQUAL_TO = '&#8807'
LESS_THAN_BUT_NOT_EQUAL_TO = '&#8808'
GREATER_THAN_BUT_NOT_EQUAL_TO = '&#8809'
MUCH_LESS_THAN = '&#8810'
MUCH_GREATER_THAN = '&#8811'
BETWEEN = '&#8812'
NOT_EQUIVALENT_TO = '&#8813'
NOT_LESS_THAN = '&#8814'
NOT_GREATER_THAN = '&#8815'
NEITHER_LESS_THAN_NOR_EQUAL_TO = '&#8816'
NEITHER_GREATER_THAN_NOR_EQUAL_TO = '&#8817'
LESS_THAN_OR_EQUIVALENT_TO = '&#8818'
GREATER_THAN_OR_EQUIVALENT_TO = '&#8819'
NEITHER_LESS_THAN_NOR_EQUIVALENT_TO = '&#8820'
NEITHER_GREATER_THAN_NOR_EQUIVALENT_TO = '&#8821'
LESS_THAN_OR_GREATER_THAN = '&#8822'
GREATER_THAN_OR_LESS_THAN = '&#8823'
NEITHER_LESS_THAN_NOR_GREATER_THAN = '&#8824'
NEITHER_GREATER_THAN_NOR_LESS_THAN = '&#8825'
PRECEDES = '&#8826'
SUCCEEDS = '&#8827'
PRECEDES_OR_EQUAL_TO = '&#8828'
SUCCEEDS_OR_EQUAL_TO = '&#8829'
PRECEDES_OR_EQUIVALENT_TO = '&#8830'
SUCCEEDS_OR_EQUIVALENT_TO = '&#8831'
DOES_NOT_PRECEDE = '&#8832'
DOES_NOT_SUCCEED = '&#8833'
SUBSET_OF = '&#8834'
SUPERSET_OF = '&#8835'
NOT_A_SUBSET_OF = '&#8836'
NOT_A_SUPERSET_OF = '&#8837'
SUBSET_OF_OR_EQUAL_TO = '&#8838'
SUPERSET_OF_OR_EQUAL_TO = '&#8839'
NEITHER_A_SUBSET_OF_NOR_EQUAL_TO = '&#8840'
NEITHER_A_SUPERSET_OF_NOR_EQUAL_TO = '&#8841'
SUBSET_OF_WITH_NOT_EQUAL_TO = '&#8842'
SUPERSET_OF_WITH_NOT_EQUAL_TO = '&#8843'
MULTISET = '&#8844'
MULTISET_MULTIPLICATION = '&#8845'
MULTISET_UNION = '&#8846'
SQUARE_IMAGE_OF = '&#8847'
SQUARE_ORIGINAL_OF = '&#8848'
SQUARE_IMAGE_OF_OR_EQUAL_TO = '&#8849'
SQUARE_ORIGINAL_OF_OR_EQUAL_TO = '&#8850'
SQUARE_CAP = '&#8851'
SQUARE_CUP = '&#8852'
CIRCLED_PLUS = '&#8853'
CIRCLED_MINUS = '&#8854'
CIRCLED_TIMES = '&#8855'
CIRCLED_DIVISION_SLASH = '&#8856'
CIRCLED_DOT_OPERATOR = '&#8857'
CIRCLED_RING_OPERATOR = '&#8858'
CIRCLED_ASTERISK_OPERATOR = '&#8859'
CIRCLED_EQUALS = '&#8860'
CIRCLED_DASH = '&#8861'
SQUARED_PLUS = '&#8862'
SQUARED_MINUS = '&#8863'
SQUARED_TIMES = '&#8864'
SQUARED_DOT_OPERATOR = '&#8865'
RIGHT_TACK = '&#8866'
LEFT_TACK = '&#8867'
DOWN_TACK = '&#8868'
UP_TACK = '&#8869'
ASSERTION = '&#8870'
MODELS = '&#8871'
TRUE = '&#8872'
FORCES = '&#8873'
TRIPLE_VERTICAL_BAR_RIGHT_TURNSTILE = '&#8874'
DOUBLE_VERTICAL_BAR_DOUBLE_RIGHT_TURNSTILE = '&#8875'
DOES_NOT_PROVE = '&#8876'
NOT_TRUE = '&#8877'
DOES_NOT_FORCE = '&#8878'
NEGATED_DOUBLE_VERTICAL_BAR_DOUBLE_RIGHT_TURNSTILE = '&#8879'
PRECEDES_UNDER_RELATION = '&#8880'
SUCCEEDS_UNDER_RELATION = '&#8881'
NORMAL_SUBGROUP_OF = '&#8882'
CONTAINS_AS_NORMAL_SUBGROUP = '&#8883'
NORMAL_SUBGROUP_OF_OR_EQUAL_TO = '&#8884'
CONTAINS_AS_NORMAL_SUBGROUP_OR_EQUAL_TO = '&#8885'
ORIGINAL_OF = '&#8886'
IMAGE_OF = '&#8887'
MULTIMAP = '&#8888'
HERMITIAN_CONJUGATE_MATRIX = '&#8889'
INTERCALATE = '&#8890'
XOR = '&#8891'
NAND = '&#8892'
NOR = '&#8893'
RIGHT_ANGLE_WITH_ARC = '&#8894'
RIGHT_TRIANGLE = '&#8895'
N_ARY_LOGICAL_AND = '&#8896'
N_ARY_LOGICAL_OR = '&#8897'
N_ARY_INTERSECTION = '&#8898'
N_ARY_UNION = '&#8899'
DIAMOND_OPERATOR = '&#8900'
DOT_OPERATOR = '&#8901'
STAR_OPERATOR = '&#8902'
DIVISION_TIMES = '&#8903'
BOWTIE = '&#8904'
LEFT_NORMAL_FACTOR_SEMIDIRECT_PRODUCT = '&#8905'
RIGHT_NORMAL_FACTOR_SEMIDIRECT_PRODUCT = '&#8906'
LEFT_SEMIDIRECT_PRODUCT = '&#8907'
RIGHT_SEMIDIRECT_PRODUCT = '&#8908'
REVERSED_TILDE_EQUALS = '&#8909'
CURLY_LOGICAL_OR = '&#8910'
CURLY_LOGICAL_AND = '&#8911'
DOUBLE_SUBSET = '&#8912'
DOUBLE_SUPERSET = '&#8913'
DOUBLE_INTERSECTION = '&#8914'
DOUBLE_UNION = '&#8915'
PITCHFORK = '&#8916'
EQUAL_AND_PARALLEL_TO = '&#8917'
LESS_THAN_WITH_DOT = '&#8918'
GREATER_THAN_WITH_DOT = '&#8919'
VERY_MUCH_LESS_THAN = '&#8920'
VERY_MUCH_GREATER_THAN = '&#8921'
LESS_THAN_EQUAL_TO_OR_GREATER_THAN = '&#8922'
GREATER_THAN_EQUAL_TO_OR_LESS_THAN = '&#8923'
EQUAL_TO_OR_LESS_THAN = '&#8924'
EQUAL_TO_OR_GREATER_THAN = '&#8925'
EQUAL_TO_OR_PRECEDES = '&#8926'
EQUAL_TO_OR_SUCCEEDS = '&#8927'
DOES_NOT_PRECEDE_OR_EQUAL = '&#8928'
DOES_NOT_SUCCEED_OR_EQUAL = '&#8929'
NOT_SQUARE_IMAGE_OF_OR_EQUAL_TO = '&#8930'
NOT_SQUARE_ORIGINAL_OF_OR_EQUAL_TO = '&#8931'
SQUARE_IMAGE_OF_OR_NOT_EQUAL_TO = '&#8932'
SQUARE_ORIGINAL_OF_OR_NOT_EQUAL_TO = '&#8933'
LESS_THAN_BUT_NOT_EQUIVALENT_TO = '&#8934'
GREATER_THAN_BUT_NOT_EQUIVALENT_TO = '&#8935'
PRECEDES_BUT_NOT_EQUIVALENT_TO = '&#8936'
SUCCEEDS_BUT_NOT_EQUIVALENT_TO = '&#8937'
NOT_NORMAL_SUBGROUP_OF = '&#8938'
DOES_NOT_CONTAIN_AS_NORMAL_SUBGROUP = '&#8939'
NOT_NORMAL_SUBGROUP_OF_OR_EQUAL_TO = '&#8940'
DOES_NOT_CONTAIN_AS_NORMAL_SUBGROUP_OR_EQUAL = '&#8941'
VERTICAL_ELLIPSIS = '&#8942'
MIDLINE_HORIZONTAL_ELLIPSIS = '&#8943'
UP_RIGHT_DIAGONAL_ELLIPSIS = '&#8944'
DOWN_RIGHT_DIAGONAL_ELLIPSIS = '&#8945'
ELEMENT_OF_WITH_LONG_HORIZONTAL_STROKE = '&#8946'
ELEMENT_OF_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = '&#8947'
SMALL_ELEMENT_OF_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = '&#8948'
ELEMENT_OF_WITH_DOT_ABOVE = '&#8949'
ELEMENT_OF_WITH_OVERBAR = '&#8950'
SMALL_ELEMENT_OF_WITH_OVERBAR = '&#8951'
ELEMENT_OF_WITH_UNDERBAR = '&#8952'
ELEMENT_OF_WITH_TWO_HORIZONTAL_STROKES = '&#8953'
CONTAINS_WITH_LONG_HORIZONTAL_STROKE = '&#8954'
CONTAINS_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = '&#8955'
SMALL_CONTAINS_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = '&#8956'
CONTAINS_WITH_OVERBAR = '&#8957'
SMALL_CONTAINS_WITH_OVERBAR = '&#8958'
Z_NOTATION_BAG_MEMBERSHIP = '&#8959'



from epyk.core.js.packages import JsPackage
from epyk.core.js.primitives import JsObjects
from epyk.core.js import JsUtils
from epyk.core.js.html import JsHtml


class Radio(JsPackage):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.varName, self.varData, self.__var_def = varName, "", None
    self._src, self._report = htmlObj, report
    self._js, self._jquery = [], None

  def check(self):
    """
    Description:
    -----------

    """
    return JsObjects.JsObjects.get("%s.checked = true" % self._src.input.dom.varName)

  def uncheck(self):
    """
    Description:
    -----------

    """
    return JsObjects.JsObjects.get("%s.checked = false" % self._src.input.dom.varName)


class CheckButton(JsPackage):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.varName, self.varData, self.__var_def = varName, "", None
    self._src, self._report = htmlObj, report
    self._js, self._jquery = [], None

  def checked(self):
    """
    Description:
    -----------

    """
    return JsObjects.JsObjects.get("%s.querySelector('i').classList.replace('fa-times', 'fa-check'); %s.querySelector('i').style.color = '%s'" % (self._src.dom.varName, self._src.dom.varName, self._report.theme.success[1]))

  def unchecked(self):
    """
    Description:
    -----------

    """
    return JsObjects.JsObjects.get("%s.querySelector('i').classList.replace('fa-check', 'fa-times'); %s.querySelector('i').style.color = '%s'" % (self._src.dom.varName, self._src.dom.varName, self._report.theme.danger[1]))


class Menu(JsPackage):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.varName, self.varData, self.__var_def = varName, "", None
    self._src, self._report = htmlObj, report
    self._js, self._jquery = [], None

  @property
  def content(self):
    """
    Description:
    -----------
    Return the content of the component
    """
    return JsHtml.ContentFormatters(self._report, "%s.innerHTML" % self.varName)

  def set_text(self, value):
    """
    Description:
    -----------
    Set the text of the component

    Attributes:
    ----------
    :param value: String. The text to be set
    """
    value = JsUtils.jsConvertData(value, None)
    return JsObjects.JsObjects.get("%s.innerHTML = %s" % (self.varName, value))

  def set_url(self, value, target='_blank'):
    """
    Description:
    -----------
    Set the URL information of the item component

    Usage::

      item.js.set_url("https://stackoverflow.com")

    Related Pages:

      https://www.w3schools.com/tags/att_a_href.asp

    Attributes:
    ----------
    :param value: String. The url link
    :param target: String.  Teh target mode of the link component
    """
    value = JsUtils.jsConvertData(value, None)
    target = JsUtils.jsConvertData(target, None)
    return JsObjects.JsObjects.get("%s.href = %s; %s.setAttribute('target', %s)" % (self.varName, value, self.varName, target))


class Switch(JsPackage):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.htmlCode = varName if varName is not None else htmlObj.htmlCode
    self.varName, self.varData, self.__var_def = "document.getElementById('%s')" % self.htmlCode, "", None
    self._src, self._report = htmlObj, report
    self._js, self._jquery = [], None

  def toggle(self):
    """
    Description:
    -----------
    Simulate a click event on the component
    """
    return JsObjects.JsObjects.get(self._src.switch.click())

  def false(self):
    """
    Description:
    -----------
    Set the switch component to False
    """
    return JsObjects.JsObjects.get("%s.querySelector('input').checked = false; %s.querySelector('p').innerHTML = %s_data.off" % (self.varName, self.varName, self.htmlCode))

  def true(self):
    """
    Description:
    -----------
    Set the switch component to True
    """
    return JsObjects.JsObjects.get("%s.querySelector('input').checked = true; %s.querySelector('p').innerHTML = %s_data.on" % (self.varName, self.varName, self.htmlCode))


class Alerts(JsPackage):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.htmlCode = varName if varName is not None else htmlObj.htmlCode
    self.varName, self.varData, self.__var_def = "document.getElementById('%s')" % self.htmlCode, "", None
    self._src, self._report = htmlObj, report
    self._js, self._jquery = [], None

  def replay(self, time=None):
    """

    :return:
    """
    time = time or self._src.options.time
    return JsObjects.JsVoid('''
    var s = %(varName)s.style; s.opacity = 1; %(varName)s.style.display = 'block';
    (function fade(){(s.opacity-=.1)<0?s.display="none":setTimeout(fade, %(time)s)})();
    ''' % {'varName': self.varName, 'time': time})


class News(JsPackage):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.htmlCode = varName if varName is not None else htmlObj.htmlCode
    self.varName, self.varData, self.__var_def = "document.getElementById('%s')" % self.htmlCode, "", None
    self._src, self._report = htmlObj, report
    self._js, self._jquery = [], None

  def reset(self):
    """

    :return:
    """
    return JsObjects.JsVoid('''
      %(varName)s.innerHTML = '';  
      ''' % {'varName': self.varName})


class Chat(JsPackage):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.htmlCode = varName if varName is not None else htmlObj.htmlCode
    self.varName, self.varData, self.__var_def = "document.getElementById('%s')" % self.htmlCode, "", None
    self._src, self._report = htmlObj, report
    self._js, self._jquery = [], None

  def add(self, message):
    return JsObjects.JsVoid('''
      %(builder)s; %(counter)s
      ''' % {"builder": self._src.build(message), 'counter': self._src.dom.querySelector(' [name=count]').innerText(1, append=True, valType=int).r})


class Room(JsPackage):

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.htmlCode = varName if varName is not None else htmlObj.htmlCode
    self.varName, self.varData, self.__var_def = "document.getElementById('%s')" % self.htmlCode, "", None
    self._src, self._report = htmlObj, report
    self._js, self._jquery = [], None

  def typing(self):
    """

    """
    return self._src.dom.querySelector("div[name=dots]").show(duration=3000)





"""
CSS Style module for the Popup components
"""

from epyk.core.css.styles.classes import CssStyle


class CssPopupTable(CssStyle.Style):
  _attrs = {'border-spacing': 0, 'border-collapse': 'collapse', 'margin': 0, 'padding': 0, 'width': '100%'}
  _selectors = {'child': 'table'}

  def customize(self):
    self.css({"background": self.rptObj.theme.colors[0], "color": self.rptObj.theme.greys[-1],
              'border': '1px solid %s' % self.rptObj.theme.colors[0],
              "box-shadow": '0 0 1px 1px %s' % self.rptObj.theme.colors[-1],
              "-webkit-box-shadow": '0 0 1px 1px %s' % self.rptObj.theme.colors[-1],
              "-moz-box-shadow": '0 0 1px 1px %s' % self.rptObj.theme.colors[-1]})


class CssPopupTableTitle(CssStyle.Style):
  _attrs = {'margin': 0, 'padding': 0, 'border-collapse': 'collapse', 'cursor': 'pointer'}
  _selectors = {'child': 'table tr:first-child th'}

  def customize(self):
    self.css({"background-color": self.rptObj.theme.colors[-1], "border": "1px solid %s" % self.rptObj.theme.colors[-1],
              "color": self.rptObj.theme.greys[0]})


class CssPopupTableTitleContent(CssStyle.Style):
  _attrs = {'vertical-align': 'middle', 'border-collapse': 'collapse', 'text-align': 'right', 'font-weight': 'bold',
            'text-transform': 'uppercase'}
  _selectors = {'child': 'table th'}

  def customize(self):
    self.css({"background": self.rptObj.theme.colors[0], "color": self.rptObj.theme.greys[-1],
                  'border': '1px solid %s' % self.rptObj.theme.colors[0]})


class CssEventLoading(CssStyle.Style):
  _attrs = {'bottom': '5px', 'right': '20px', 'position': 'fixed', 'padding': '5px'}

  def customize(self):
    self.css({"background-color": self.rptObj.theme.greys[0], 'color': self.rptObj.theme.greys[-1]})



"""
Module in charge of the CSS Standard modules
"""

from epyk.core.css.styles.classes import CssStyle
from epyk.core.css import Defaults


class CssBody(CssStyle.Style):
  _attrs = {'top': 0, 'margin': '0 20px 0 20px'}
  cssId = {'tag': 'body'}

  def customize(self):
    if isinstance(Defaults.BACKGROUND, tuple):
      bg_color = self.rptObj.theme.greys[0]
      if bg_color != '#000000':
        bg_color = self.rptObj.theme.greys[2]
    else:
      bg_color = Defaults.BACKGROUND
    self.css({"background-color": bg_color, "color": self.rptObj.theme.greys[9], 'font-family':
      Defaults.Font.family, 'font-size': "%s%s" % (Defaults.Font.size, Defaults.Font.unit)})


class CssTextSelection(CssStyle.Style):
  def customize(self):
    self.css({"background": self.rptObj.theme.success[0]})


class CssBodyContent(CssStyle.Style):
  _attrs = {'margin-top': '10px', 'padding': '5px'}

  def customize(self):
    self.css({"background-color": self.rptObj.theme.greys[0], "border-radius": "5px",
              "border": '1px solid %s' % self.rptObj.theme.greys[3]})
    if Defaults.BODY_CONTAINER is not None:
      self.css(Defaults.BODY_CONTAINER)


class CssBodyLoadingBack(CssStyle.Style):
  _attrs = {'text-align': 'center', 'top': 0, 'left': 0, 'width': '100%', 'padding-top': '20%', 'height': '100%',
            'z-index': 295, 'position': 'fixed', 'opacity': 0.5, 'filter': 'alpha(opacity=50)'}

  def customize(self):
    self.css({"background-color": self.rptObj.theme.greys[5]})


class CssBodyLoading(CssStyle.Style):
  _attrs = {'text-align': 'center', 'top': 0, 'left': 0, 'width': '100%', 'position': 'fixed', 'padding-top': '10%',
            'height': '100%', 'display': 'none', 'z-index': 300}

  def customize(self):
    self.css({"color": self.rptObj.theme.greys[9]})


class CssNotSelect(CssStyle.Style):
  _attrs = {"-webkit-touch-callout": 'none', "user-select": 'none', "-webkit-user-select": 'none',
            "-khtml-user-select": 'none', "-moz-user-select": 'none', "-ms-user-select": 'none'}

class CssCloseSpan(CssStyle.Style):
  _attrs = {'float': 'right', 'text-align': 'right', 'font-size': '32px', 'z-index': 10, 'color': 'red',
            'position': 'relative', 'top': '-10px'}



from epyk.core.data.DataClass import DataClass


class FixedColumns(DataClass):

  def activate(self):
    """
    Description:
    -----------

    Related Pages:

			https://datatables.net/reference/option/fixedColumns
    """
    self.leftColumns = 2
    return self

  @property
  def heightMatch(self):
    """
    Description:
    -----------
    FixedColumns operates by cloning the cells from the original table and then inserting them into the document, positioned visually above the DataTable - thus allowing them to appear fixed.

    Related Pages:

			https://datatables.net/reference/option/fixedColumns.heightMatch
    """
    return self._attrs["heightMatch"]

  @heightMatch.setter
  def heightMatch(self, val):
    self._attrs["heightMatch"] = val

  @property
  def leftColumns(self):
    return self._attrs["leftColumns"]


  @leftColumns.setter
  def leftColumns(self, val):
    self._attrs["leftColumns"] = val

  @property
  def rightColumns(self):
    return self._attrs["rightColumns"]

  @rightColumns.setter
  def rightColumns(self, val):
    self._attrs["rightColumns"] = val




from epyk.core.js import JsUtils
from epyk.core.js.packages import JsPackage
from epyk.core.html.options import OptSortable
from epyk.core.js.primitives import JsObjects


class Sortable(JsPackage):

  lib_alias = {"js": 'sortable'}
  lib_set_var = True

  def __init__(self, src=None, varName=None, selector=None, setVar=True, parent=None):
    super(Sortable, self).__init__(src=src, varName=varName, selector=selector, data=None, setVar=setVar, parent=parent)
    self.__options = OptSortable.OptionsSortable(self)

  @property
  def options(self):
    """
    Description:
    ------------
    Sortable options

    Related Pages:

			https://github.com/SortableJS/Sortable

		:rtype: OptSortable.OptionsSortable
    """
    return self.__options

  def create(self, htmlElement, options=None):
    """
    Description:
    ------------
    Create new instance

    Related Pages:

			https://github.com/SortableJS/Sortable

    Attributes:
    ----------
    :param dom:
    :param options: Optional. Dictionary with the sortable options
    """
    self.__options.update(options)
    return self

  def toArray(self):
    """
    Description:
    ------------
    Serializes the sortable's item data-id's (dataIdAttr option) into an array of string.

    Related Pages:

			https://github.com/SortableJS/toArray
    """
    return self.fnc_closure("toArray()")

  def sort(self, data):
    """
    Description:
    ------------
    Sorts the elements according to the array.

    Related Pages:

			https://github.com/SortableJS/Sortable

    Attributes:
    ----------
    :param data:
    """
    data = JsUtils.jsConvertData(data, None)
    return self.fnc_closure("sort(%s)" % data)

  def save(self):
    """
    Description:
    ------------
    Save the current sorting

    Related Pages:

			https://github.com/SortableJS/save
    """
    return self.fnc_closure("destroy()")

  def destroy(self):
    """
    Description:
    ------------
    Removes the sortable functionality completely.

    Related Pages:

			https://github.com/SortableJS/destroy
    """
    return self.fnc_closure("destroy()")

  def toStr(self):
    if self.setVar:
      self._selector = "Sortable.create(%s, %s)" % (self._selector, JsUtils.jsConvertData(self.__options, None))
    return super(Sortable, self).toStr()




from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects


class JsError(object):

  def __init__(self, val):
    self.__selector = val

  @property
  def message(self):
    """
    Description:
    ------------
    Sets or returns an error message (a string)
    """
    return JsObjects.JsObject.JsObject.get("%s.message" % self.__selector)

  @property
  def name(self):
    """
    Description:
    ------------
    Sets or returns an error name
    """
    return JsObjects.JsObject.JsObject.get("%s.name" % self.__selector)

  def toStr(self):
    return JsObjects.JsObject.JsObject.get(self.__selector)

  def __str__(self):
    return self.__selector


class JsTry(object):

  def __init__(self, jsFncs):
    self.__try_jsFnc = jsFncs
    self.__fin_jsFnc = None
    self.__catch_jsFnc = None

  @property
  def error(self):
    """
    Description:
    ------------
    The error message object
    """
    return JsError("err")

  def catch(self, jsFncs):
    """
    Description:
    ------------
    Block of code to handle errors

    The variable to be used is err in this loop.

    Attributes:
    ----------
    :param jsFncs: List or String. The JavaScript logic
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.__catch_jsFnc = jsFncs
    return self

  def except_(self, jsFncs):
    """
    Description:
    ------------
    Block of code to handle errors (catch in JavaScript)

    The variable to be used is err in this loop.

    Attributes:
    ----------
    :param jsFncs: List or String. The JavaScript logic
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.__catch_jsFnc = jsFncs
    return self

  def finally_(self, jsFncs):
    """
    Description:
    ------------
    Block of code to be executed regardless of the try / catch result

    Attributes:
    ----------
    :param jsFncs: List or String. The JavaScript logic
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.__fin_jsFnc = jsFncs
    return self

  def toStr(self):
    """
    Description:
    ------------

    """
    a = JsUtils.jsConvertFncs(self.__try_jsFnc, toStr=True)
    if self.__catch_jsFnc is None:
      raise Exception("Catch must be defined")

    b = JsUtils.jsConvertFncs(self.__catch_jsFnc, toStr=True)
    if self.__fin_jsFnc is None:
      return "try{%s} catch(%s){%s}" % (a, self.error, b)

    c = JsUtils.jsConvertFncs(self.__fin_jsFnc, toStr=True)
    return "try{%s} catch(%s){%s} finally{%s}" % (a, self.error, b, c)

  def __str__(self):
    return self.toStr()





from epyk.core.html import Html
from epyk.core.js import JsUtils


class Chart(Html.Html):
  name = 'Google Chart'
  requirements = ('google-charts', )

  def __init__(self,  report, data, width, height, htmlCode, options, profile):
    self.height = height[0]
    super(Chart, self).__init__(report, data, htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self._d3, self._chart, self._datasets, self._options, self._data_attrs, self._attrs = None, None, [], None, {}, {}
    self.__options = options
    self.style.css.margin_top = 10

  @property
  def chartId(self):
    """
    Description:
    -----------
    Return the Javascript variable of the chart
    """
    return "%s_obj" % self.htmlCode

  def build(self, data=None, options=None, profile=False):
    return '''
      %(chartId)s = google.charts.setOnLoadCallback( (function(){
        var data = new google.visualization.DataTable();
        var chartData = %(data)s;
        data.addColumn('string', chartData.x);
        chartData.series.forEach(function(col){data.addColumn('number', col)})
        data.addRows(chartData.datasets);
        
        var chart = new google.visualization.%(type)s(%(varId)s);
        chart.draw(data, {});
        return chart
      }));
      ''' % {'chartId': self.chartId, 'varId': self.dom.varId, 'data': JsUtils.jsConvertData(data, None), 'type': self.__options["type"]}

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>' % self.get_attrs(pyClassNames=self.style.get_classes())


class ChartLine(Chart):
  pass



"""
Documentation:
    - https://www.html5rocks.com/en/tutorials/file/filesystem/
    - https://w3c.github.io/filesystem-api/
    - https://blog.teamtreehouse.com/building-an-html5-text-editor-with-the-filesystem-apis
"""

from epyk.core.js import JsUtils


class JsFileSystem(object):
  class __internal(object):
    _context = {}

  def __init__(self, src=None):
    self.src = src if src is not None else self.__internal()
    self._js = []

  def toStr(self):
    """

    :return:
    """
    return '''
      if (window.webkitRequestFileSystem !== undefined){
        console.log(window.webkitRequestFileSystem);
        window.webkitRequestFileSystem(window.TEMPORARY, 1024*1024, function(fs) {
          console.log(fs);
          fs.root.getFile('log.txt', {create: true}, function(fileEntry) {
              console.log(fileEntry);
          })
        })
      }'''




from epyk.core.data.DataClass import DataClass

from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects


class Annotations(DataClass):

  @property
  def drawTime(self):
    """
    Description:
    -----------
    Overrides annotation.drawTime if set

    Related Pages:

      https://github.com/chartjs/chartjs-plugin-annotation
    """
    return self.get('afterDraw')

  @drawTime.setter
  def drawTime(self, value):
    self.set(value)

  @property
  def type(self):
    """
    Description:
    -----------

    Related Pages:

      https://github.com/chartjs/chartjs-plugin-annotation
    """
    return self.get('line')

  @type.setter
  def type(self, value):
    self.set(value)

  @property
  def mode(self):
    """
    Description:
    -----------

    Related Pages:

      https://github.com/chartjs/chartjs-plugin-annotation
    """
    return self.get('horizontal')

  @mode.setter
  def mode(self, value):
    self.set(value)

  @property
  def scaleID(self):
    """
    Description:
    -----------

    Related Pages:

      https://github.com/chartjs/chartjs-plugin-annotation
    """
    return self.get('y-axis-0')

  @scaleID.setter
  def scaleID(self, value):
    self.set(value)

  @property
  def value(self):
    """
    Description:
    -----------

    Related Pages:

      https://github.com/chartjs/chartjs-plugin-annotation
    """
    return self.get('25')

  @value.setter
  def value(self, value):
    self.set(value)

  @property
  def borderColor(self):
    """
    Description:
    -----------

    Related Pages:

      https://github.com/chartjs/chartjs-plugin-annotation
    """
    return self.get('red')

  @borderColor.setter
  def borderColor(self, color):
    self.set(color)

  @property
  def borderWidth(self):
    """
    Description:
    -----------

    Related Pages:

      https://github.com/chartjs/chartjs-plugin-annotation
    """
    return self.get(2)

  @borderWidth.setter
  def borderWidth(self, num):
    self.set(num)

  def onClick(self, jsFncs):
    """
    Description:
    -----------
    Fires when the user clicks this annotation on the chart
    (be sure to enable the event in the events array below).
    
    Related Pages:

      https://github.com/chartjs/chartjs-plugin-annotation

    Attributes:
    ----------
    :param jsFncs:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._attrs["onClick"] = JsObjects.JsVoid(
      "function(event) { %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))


class Annotation(DataClass):

  @property
  def drawTime(self):
    """
    Description:
    -----------

    Related Pages:

      https://github.com/chartjs/chartjs-plugin-annotation
    """
    return self.get('afterDatasetsDraw')

  @drawTime.setter
  def drawTime(self, value):
    self.set(value)

  @property
  def events(self):
    """
    Description:
    -----------

    Related Pages:

      https://github.com/chartjs/chartjs-plugin-annotation
    """
    return self.get(['click'])

  @events.setter
  def events(self, value):
    self.set(value)

  @property
  def dblClickSpeed(self):
    """
    Description:
    -----------

    Related Pages:

      https://github.com/chartjs/chartjs-plugin-annotation
    """
    return self.get(350)

  @dblClickSpeed.setter
  def dblClickSpeed(self, value):
    self.set(value)

  @property
  def annotations(self):
    """
    Description:
    -----------
    The plugin allows for horizontal zooming by clicking and dragging over the chart.

    Related Pages:

      https://chartjs-plugin-crosshair.netlify.app/options.html
    """
    return self.sub_data("annotations", Annotations)




from epyk.core.html.graph import GraphChartJs


class Choropleth(GraphChartJs.Chart):
  name = 'ChartJs Choropleth'
  requirements = ('chartjs-chart-geo', )

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(Choropleth, self).__init__(report, width, height, htmlCode, options, profile)
    self._attrs['type'] = 'choropleth'




"""
Create a HTML page only from JavaScript functions
"""

from epyk.core.js import Js
from epyk.core.js import JsUtils
from epyk.tests import test_statics

# Create a blank Javascript Python object
jsObj = Js.JsBase()
dom = jsObj.createElement("div", "testDom")


# Write the results to a Javascript file with a Launcher
f = JsUtils.JsFile("TestJquery", path=test_statics.OUTPUT_PATHS)

# Write the Javascript fragments to the file
f.writeJs([
  dom.attr("id", "jq"),
  jsObj.body.appendChild(dom),
  # Use Jquery from the new dom object
  dom.jquery.text("test").css("color", "red").html("<b>test</b>test"),
  dom.jquery.click(jsObj.console.log(
    jsObj.objects.jqThis.html()
  ))
])

# Close the file and print the location of the launcher
print(f.close(jsObj))



from epyk.core.html import geo


class Plotly(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Plotly"

  @property
  def choropleths(self):
    """
    Description:
    -----------
    Plotly Choropleth charts

    Related Pages:

			https://plotly.com/javascript/mapbox-county-choropleth/
    """
    return PlotlyChoropleth(self.parent)

  @property
  def bubbles(self):
    """
    Description:
    -----------
    Plotly Bubble charts

    Related Pages:

			https://plotly.com/javascript/bubble-maps/
    """
    return PlotlyBubble(self.parent)

  def scattermapbox(self, record, lon_columns=None, lat_columns=None, text_columns=None,
                    profile=None, options=None,  width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Scatter`

    Related Pages:

      https://plot.ly/javascript/mapbox-layers/

    Attributes:
    ----------
    :param record:
    :param lon_columns:
    :param lat_columns:
    :param text_columns:
    :param title:
    :param filters:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    series = []
    for i, l in enumerate(lon_columns):
      series.append({'lon': [], 'lat': [], 'text': []})
      for rec in record:
        series[-1]['lon'].append(rec.get(l, 0))
        series[-1]['lat'].append(rec.get(lat_columns[i], 0))
        if text_columns is not None:
          series[-1]['text'].append(rec.get(text_columns[i], 0))
    line_chart = geo.GeoPlotly.Scatter(self.parent.context.rptObj, width, height, options or {}, htmlCode,
                                      profile)
    line_chart.options.responsive = True
    for i, s in enumerate(series):
      line_chart.add_trace(s)
      line_chart.data.marker.color = self.parent.context.rptObj.theme.colors[::-1][i]
    line_chart.layout.mapbox.style = "open-street-map"
    # line_chart.data.marker.size = 4
    # line_chart.layout.mapbox.style = "open-street-map"
    # line_chart.layout.mapbox.center.lat = 38
    # line_chart.layout.mapbox.center.lon = -90
    # line_chart.layout.dragmode = 'zoom'
    # line_chart.layout.mapbox.zoom = 3
    return line_chart

  def density(self, record, y_columns=None, x_axis=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Scatter`

    Related Pages:

      https://plot.ly/javascript/mapbox-density-heatmaps/

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    line_chart = geo.GeoPlotly.Scatter(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    line_chart.options.responsive = True
    line_chart.add_trace({"lon": [-112.8352], 'lat': [48.4113], 'z': [20], 'text': [0.0875]}, type="densitymapbox")
    line_chart.data.marker.color = "fuchsia"
    line_chart.data.marker.size = 4
    line_chart.layout.mapbox.style = "stamen-terrain"
    line_chart.layout.mapbox.center.lat = 38
    line_chart.layout.mapbox.center.lon = -90
    line_chart.layout.dragmode = 'zoom'
    line_chart.layout.zoom = 3
    return line_chart

  def chorolet(self, record, y_columns=None, x_axis=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Chorolet`

    Related Pages:

      https://plotly.com/javascript/mapbox-county-choropleth/

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    data = {"locations": ["NY", "MA", "VT"], "z": [-50, -10, -20]}
    line_chart = geo.GeoPlotly.Chorolet(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    line_chart.options.responsive = True
    data['geojson'] = "https://raw.githubusercontent.com/python-visualization/folium/master/examples/data/us-states.json"
    line_chart.add_trace(data)
    line_chart.layout.mapbox.center.lat = 38
    line_chart.layout.mapbox.center.lon = -90
    line_chart.layout.dragmode = 'zoom'
    line_chart.layout.zoom = 3
    return line_chart


class PlotlyBubble(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Plotly"

  def bubble(self, scope, record=None, size_col=None, country_col=None, long_col=None, lat_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    How to make a D3.js-based bubble map in JavaScript. A bubble map overlays a bubble chart on a map.

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.BubbleGeo`

    Related Pages:

      https://plotly.com/javascript/bubble-maps/

    Attributes:
    ----------
    :param scope: String. The scope of the chart
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param long_col: String. The column in the recordset used to retrieve the longitude (optional if country codes)
    :param lat_col: String. The column in the recordset used to retrieve the latitude (optional if country codes)
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    dftl_options = {'type': 'scattergeo', 'mode': 'markers'}
    if options is not None:
      dftl_options.update(options)
    map_chart = geo.GeoPlotly.BubbleGeo(self.parent.context.rptObj, width, height, dftl_options, htmlCode, profile)
    map_chart.options.responsive = True

    records = []
    if country_col is not None:
      records = self.parent.context.rptObj.data.plotly.countries(record, country_col, size_col, dftl_options.get('scale', False))
      points = ['locations']
    elif long_col is not None and lat_col is not None:
      records = self.parent.context.rptObj.data.plotly.locations(record, long_col, lat_col, size_col, dftl_options.get('scale', False))
      points = ['lon', 'lat']
    for record in records:
      map_chart.add_trace({p: record[p] for p in points})
      #map_chart.data.marker.colorbar.title = "Test"
      #map_chart.data.marker.line.color = "black"
      map_chart.data.marker.size = record['marker']['size']
      # map_chart.data.marker.cmin = 0
      # map_chart.data.marker.cmax = max(values)
      map_chart.data.marker.color = record['marker']['size']
      map_chart.data.marker.colorscale = 'Reds'
    map_chart.layout.geo.scope = scope
    map_chart.layout.geo.resolution = 150
    return map_chart

  def usa(self, record=None, size_col=None, country_col=None, long_col=None, lat_col=None, text_columns=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A bubble chart for the USA.

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.BubbleGeo`

    Related Pages:

      https://plotly.com/javascript/bubble-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param long_col: String. The column in the recordset used to retrieve the longitude (optional if country codes)
    :param lat_col: String. The column in the recordset used to retrieve the latitude (optional if country codes)
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    map_usa = self.bubble('usa', record, size_col=size_col, country_col=country_col, long_col=long_col, lat_col=lat_col, profile=profile,
                          options=options, width=width, height=height, htmlCode=htmlCode)
    map_usa.data.locationmode = 'USA-states'
    map_usa.data.locationmode = 'usa'
    map_usa.layout.no_background()
    return map_usa

  def europe(self, record=None, size_col=None, country_col=None, long_col=None, lat_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A bubble chart for Europe

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.BubbleGeo`

    Related Pages:

      https://plotly.com/javascript/bubble-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param long_col: String. The column in the recordset used to retrieve the longitude (optional if country codes)
    :param lat_col: String. The column in the recordset used to retrieve the latitude (optional if country codes)
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    return self.bubble('europe', record, size_col=size_col, country_col=country_col, long_col=long_col, lat_col=lat_col,
                       profile=profile, options=options, width=width, height=height, htmlCode=htmlCode)

  def asia(self, record=None, size_col=None, country_col=None, long_col=None, lat_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A bubble chart for Asia

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.BubbleGeo`

    Related Pages:

      https://plotly.com/javascript/bubble-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param long_col: String. The column in the recordset used to retrieve the longitude (optional if country codes)
    :param lat_col: String. The column in the recordset used to retrieve the latitude (optional if country codes)
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    return self.bubble('asia', record, size_col=size_col, country_col=country_col, long_col=long_col, lat_col=lat_col,
                       profile=profile, options=options, width=width, height=height, htmlCode=htmlCode)

  def africa(self, record=None, size_col=None, country_col=None, long_col=None, lat_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A bubble chart for Africa

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.BubbleGeo`

    Related Pages:

      https://plotly.com/javascript/bubble-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param long_col: String. The column in the recordset used to retrieve the longitude (optional if country codes)
    :param lat_col: String. The column in the recordset used to retrieve the latitude (optional if country codes)
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    return self.bubble('africa', record, size_col=size_col, country_col=country_col, long_col=long_col, lat_col=lat_col,
                       profile=profile, options=options, width=width, height=height, htmlCode=htmlCode)

  def south_america(self, record=None, size_col=None, country_col=None, long_col=None, lat_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A bubble chart for South America

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.BubbleGeo`

    Related Pages:

      https://plotly.com/javascript/bubble-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param long_col: String. The column in the recordset used to retrieve the longitude (optional if country codes)
    :param lat_col: String. The column in the recordset used to retrieve the latitude (optional if country codes)
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    return self.bubble('south america', record, size_col=size_col, country_col=country_col, long_col=long_col, lat_col=lat_col,
                       profile=profile, options=options, width=width, height=height, htmlCode=htmlCode)

  def north_america(self, record=None, size_col=None, country_col=None, long_col=None, lat_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A bubble chart for North America

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.BubbleGeo`

    Related Pages:

      https://plotly.com/javascript/bubble-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param long_col: String. The column in the recordset used to retrieve the longitude (optional if country codes)
    :param lat_col: String. The column in the recordset used to retrieve the latitude (optional if country codes)
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    return self.bubble('north america', record, size_col=size_col, country_col=country_col, long_col=long_col, lat_col=lat_col,
                       profile=profile, options=options, width=width, height=height, htmlCode=htmlCode)

  def world(self, record=None, size_col=None, country_col=None, long_col=None, lat_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A world bubble chart

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.BubbleGeo`

    Related Pages:

      https://plotly.com/javascript/bubble-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param long_col: String. The column in the recordset used to retrieve the longitude (optional if country codes)
    :param lat_col: String. The column in the recordset used to retrieve the latitude (optional if country codes)
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    """
    return self.bubble('world', record, size_col=size_col, country_col=country_col, long_col=long_col, lat_col=lat_col,
                       profile=profile, options=options, width=width, height=height, htmlCode=htmlCode)


class PlotlyChoropleth(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Plotly"

  def world(self, record, size_col=None, country_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A world Choropleth Chart

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    dftl_options = {'type': 'choropleth', 'mode': ''}
    if options is not None:
      dftl_options.update(options)

    records = self.parent.context.rptObj.data.plotly.choropleth(record, country_col, size_col, dftl_options.get('scale', False))
    line_chart = geo.GeoPlotly.Choropleth(self.parent.context.rptObj, width, height, dftl_options, htmlCode, profile)
    line_chart.options.responsive = True
    for record in records:
      line_chart.add_trace(record)
    return line_chart

  def europe(self, record, size_col=None, country_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A Choropleth Chart for European countries

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    map_chart = self.world(record, size_col, country_col, profile, options, width, height, htmlCode)
    map_chart.layout.geo.scope = 'europe'
    return map_chart

  def asia(self, record, size_col=None, country_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A Choropleth Chart for asian countries

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    map_chart = self.world(record, size_col, country_col, profile, options, width, height, htmlCode)
    map_chart.layout.geo.scope = 'asia'
    return map_chart

  def africa(self, record, size_col=None, country_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A Choropleth Chart for african countries

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    map_chart = self.world(record, size_col, country_col, profile, options, width, height, htmlCode)
    map_chart.layout.geo.scope = 'africa'
    return map_chart

  def south_america(self, record, size_col=None, country_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A Choropleth Chart for south american countries

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    map_chart = self.world(record, size_col, country_col, profile, options, width, height, htmlCode)
    map_chart.layout.geo.scope = 'south america'
    return map_chart

  def north_america(self, record, size_col=None, country_col=None, profile=None, options=None, width=(100, "%"), height=(430, "px"), htmlCode=None):
    """
    Description:
    -----------
    A Choropleth Chart for north american countries

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    map_chart = self.world(record, size_col, country_col, profile, options, width, height, htmlCode)
    map_chart.layout.geo.scope = 'north america'
    return map_chart

  def usa(self, record, y_column=None, x_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoPlotly.Choropleth`

    Related Pages:

      https://plotly.com/javascript/choropleth-maps/

    Attributes:
    ----------
    :param record: Data. The recordset
    :param size_col: String. The column in the recordset used for the values
    :param country_col: String. The column in the recordset used to retrieve country code
    :param profile: A flag to set the component performance storage
    :param options: Dictionary. The charts options
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode:
    """
    map_chart = geo.GeoPlotly.Choropleth(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    map_chart.options.responsive = True
    data = {}
    for rec in record:
      if x_axis in rec:
        data[rec[x_axis]] = data.get(x_axis, 0) + float(rec.get(y_column, 0))
    locations = list(data.keys())
    map_chart.add_trace({'locations': locations, 'z': [data[k] for k in locations]})
    map_chart.data.locationmode = 'USA-states'
    map_chart.layout.geo.scope = 'usa'
    map_chart.layout.geo.showlakes = True
    map_chart.layout.geo.showland = True
    return map_chart
